/**
 * HMG GraphPlus Library 2021
 * Author: S. Rathinagiri srathinagiri@gmail.com
 *
 * Following is the copyright notice for Treemap algorithm used in this Library
 * Copyright (C) 2001 by University of Maryland, College Park, MD 20742, USA 
 * and Martin Wattenberg, w@bewitched.com
 * All rights reserved.
 * Authors: Benjamin B. Bederson and Martin Wattenberg
 * http://www.cs.umd.edu/hcil/treemaps
 */



#include <hmg.ch>
#include <GraphPlus.ch>
#include "hbclass.ch"

Class GraphPlus

   Data aData          INIT {}    // Graph data
   Data aLegends       INIT {}
   Data aCategories    INIT {}  // Horizontal Axis
   Data nGraphType     INIT GT_DEFAULT  // COLUMN
   Data cTitle         INIT '' // Graph Title
   Data cXTitle        INIT '' // Horizontal Title
   Data cYTitle        INIT '' // Vertical Title
   Data lXRotate       INIT .f. // xValues Rotate
   Data nHeight        INIT 600 // Graph Height
   Data nWidth         INIT 600 // Graph Width
   Data lShowLegends   INIT .f. // Show legends
   Data lDataLabels    INIT .f. // show data labels
   Data lHGrid         INIT .t. // Horizontal Grid
   Data lVGrid         INIT .t. // Vertical Grid
   Data lShowValues    INIT .t. // Show Value Axis
   Data lShowCategories INIT .t. // show categories
   Data nLegendPos     INIT LEGEND_ON_BOTTOM   // 0 Right 1 Bottom
   Data lOutBorder     INIT .t. // draw outside border
   Data aColors        INIT {}  // aColors
   Data aTitleFont     INIT { 'Arial', 12, .t., .f., .f., .f. }
   Data aLegendFont    INIT { 'Arial', 8, .f., .f., .f., .f. }
   Data aBackColor     INIT CLR_WHITE
   Data aTitleColor    INIT { 0, 0, 255 }
   Data aTextColor      INIT { 0, 0, 0 }
   Data aGridLineColor INIT { 200, 200, 200 }
   Data nLegendBoxSize INIT 10
   Data nLegendWidth   INIT 80
   Data nColorTheme    INIT 1
   Data nUpper         INIT 0
   Data nLower         INIT 0
   Data lCategoryBelow INIT .t.
   Data nRange         INIT 0
   Data nPixelperPoint INIT 0
   Data nLineWidth     INIT 2
   Data nZeroPos       INIT 0
   Data nVDivision     INIT 5
   Data nHDivision     INIT 5
   Data nUpperDivision INIT 0
   Data nLowerDivision INIT 0
   Data cPicture       INIT ''
   Data nCategoryRow   INIT 0
   Data nScaleItemRow  INIT 0
   Data nPieGap        INIT 3
   Data nBarGapRatio   INIT 0.2 // in percentage 20%
   Data nGapWidthRatio INIT 1  // in percentage 200%
   Data nImageHeight   INIT 600
   Data nImageWidth    INIT 600
   Data nPointSize     INIT 10
   Data lGraphRotated  INIT .f.
   Data nMaxScaleWidth INIT 0
   Data cPictureX      INIT ''
   Data cPictureY      INIT ''
   Data lScatterLine   INIT .f.
   Data aSunBurstData  INIT {}
   Data nLevels        INIT 0
   Data aLevelHeaders  INIT {}
   Data aLevelValues   INIT {}
   Data aTotalItems    INIT {}
   Data nLabelHeight   INIT 0
   Data aTreeMapRect   INIT {}
   Data aTreeMapData   INIT {}
   
   Data hBitMap        INIT nil // Bitmap to store Graph
   Data hDC            INIT nil 
   Data BTStruct       INIT nil  
   Data nTop           INIT 10
   Data nLeft          INIT 10
   Data nRight         INIT 590 
   Data nBottom        INIT 590
   Data aScale         INIT {}
   DATA aColorThemes INIT { { CLR_HAVELOCK_BLUE, CLR_WESTSIDE, CLR_LIGHT_GREY, CLR_ORANGE_YELLOW, CLR_PICTON_BLUE, CLR_APPLE }, ;
      { CLR_HAVELOCK_BLUE, CLR_LIGHT_GREY, CLR_PICTON_BLUE, CLR_DARKER_BLUE, CLR_DARK_GREY, CLR_ENDEAVOUR }, ;
      { CLR_WESTSIDE, CLR_ORANGE_YELLOW, CLR_APPLE, CLR_SADDLE_BROWN, CLR_LIGHT_BROWN, CLR_DELL }, ;
      { CLR_APPLE, CLR_PICTON_BLUE, CLR_ORANGE_YELLOW, CLR_DELL, CLR_ENDEAVOUR, CLR_LIGHT_BROWN }, ;
      { CLR_HAVELOCK_BLUE }, ;
      { CLR_WESTSIDE }, ;
      }
   Data lLighter       INIT .t.
   
   Method New 
   Method Draw
   Method Bitmap
   Method DrawColumnGraph
   Method DrawBarGraph
   Method DrawLineGraph
   Method DrawPieGraph
   Method DrawPointsGraph
   Method DrawFunnelGraph
   Method DrawStackedColumnGraph
   Method DrawStackedBarGraph
   Method DrawAreaGraph
   Method DrawScatterXYGraph
   Method DrawDoughnutGraph
   Method DrawSunBurstGraph
   Method DrawSunBurstLegends   
   Method DrawWaterfallGraph
   Method DrawTreeMapGraph
   
   Method DrawLegends
   Method InitColors
   Method DrawXYTitles
   Method DrawCategories
   Method DrawVerticalScale
   Method DrawBarScale
   Method DrawHorizontalGrid
   Method DrawVerticalGrid
   Method SwitchRowsToColumns
   Method FindPieCoordinates
   Method FillSunBurstData
   Method FillTreeMapData
   
   Method StripTreemap
   Method LayoutStrip
   Method ComputeHorizontalBoxLayout
   Method ComputeSize
   Method ComputeAverageAspectRatio
   Method ComputeAspectRatio
   Method Save
   Method GraphData    SETGET
   Method GridData     SETGET
   Method Legends      SETGET
   Method Picture      SETGET
   Method PictureX     SETGET
   Method PictureY     SETGET
   Method Categories   SETGET
   Method GraphColors  SETGET
   Method GraphType    SETGET
   Method Title        SETGET
   Method XTitle       SETGET
   Method YTitle       SETGET
   Method Height       SETGET
   Method Width        SETGET
   Method ShowLegends  SETGET
   Method ShowLabels   SETGET
   Method ShowCategories SETGET
   Method ShowHGrid    SETGET
   Method ShowVGrid    SETGET
   Method ShowValues   SETGET
   Method LegendPos    SETGET
   Method ColorTheme   SETGET
   Method BackColor    SETGET
   Method BarGapRatio  SETGET
   Method BarGapWidthRatio  SETGET
   Method TitleFont    SETGET
   Method LegendFont   SETGET
   Method ScatterLine  SETGET
   Method VDivisions   SETGET
   Method HDivisions   SETGET
   Method TotalItems   SETGET
   
End class   
   
METHOD New CLASS GraphPlus
RETURN self

METHOD GraphData( aData ) CLASS GraphPlus
   local i := 0, nLen := 0
   if PCount() == 0
      return ::aData
   else
      if valtype( aData[ 1 ] ) == 'A' // 2D array
         ::aData := aData
      else
         ::aData := { aData } // 1D - make it 2D
      endif      
   endif
   return Self
   
Method GridData( cFormName, cControlName ) Class GraphPlus
   local i := 0
   local j := 0
   local nRowCount := getproperty( cFormName, cControlName, 'ITEMCOUNT' )
   local nColumnCount := 0
   local aLineData := {}
   local aLine := {}
   if PCount() == 0
      return ::aData
   else
      asize( ::aData, 0 )
      asize( ::aLegends, 0 )
      asize( ::aCategories, 0 )
      nColumnCount := getproperty( cFormName, cControlName, 'COLUMNCOUNT' )
      for i := 1 to nColumnCount
         if i == 1
//            ::cYTitle := getproperty( cFormName, cControlName, 'COLUMNHEADER', i ) not meaningful!
         else   
            aadd( ::aCategories, getproperty( cFormName, cControlName, 'COLUMNHEADER', i ) )
         endif   
      next i
      if nRowCount > 0
         for i := 1 to nRowCount
            aLineData := getproperty( cFormName, cControlName, 'ITEM', i )
            asize( aLine, 0 )
            for j := 1 to nColumnCount
               if j == 1
                  aadd( ::aLegends, aLineData[ j ] )
               else   
                  aadd( aLine, val( alltrim( aLineData[ j ] ) ) )
               endif
            next j
            aadd( ::aData, aclone( aLine ) )
         next i
      endif   
   endif
return self   

Method Bitmap( nWidth, nHeight ) Class GraphPlus
   if pcount() == 0
      return ::hBitMap   
   else   
      return BT_BitmapCopyAndResize(::hBitmap, nWidth, nHeight )
   endif   
return ::hBitMap   

METHOD GraphColors( aColors ) CLASS GraphPlus
   if PCount() == 0
      return ::aColors
   else
      ::aColors := aColors
   endif
   return Self
   
METHOD Legends( aLegends ) CLASS GraphPlus
   if PCount() == 0
      return ::aLegends
   else
      ::aLegends := aLegends
   endif
   return Self
   
METHOD Categories( aCategories ) CLASS GraphPlus
   if PCount() == 0
      return ::aCategories
   else
      ::aCategories := aCategories
   endif
   return Self

METHOD BackColor( aBackColor ) CLASS GraphPlus
   if PCount() == 0
      return ::aBackColor
   else
      ::aBackColor := aBackColor
   endif
   return Self
   
METHOD GraphType( nGraphType ) CLASS GraphPlus
   if PCount() == 0
      return ::nGraphType
   else
      ::nGraphType := nGraphType
      if ::nGraphType == 5 .or. ::nGraphType == 6 .or. ::nGraphType == 8 // bar or funnel or stacked bar
         ::lGraphRotated := .t.
      else
         ::lGraphRotated := .f.
      endif
   endif
   return Self

METHOD Title( cTitle ) CLASS GraphPlus
   if PCount() == 0
      return ::cTitle
   else
      ::cTitle := cTitle
   endif
   return Self
   
METHOD XTitle( cXTitle ) CLASS GraphPlus
   if PCount() == 0
      return ::cXTitle
   else
      ::cXTitle := cXTitle
   endif
   return Self
      
METHOD YTitle( cYTitle ) CLASS GraphPlus
   if PCount() == 0
      return ::cYTitle
   else
      ::cYTitle := cYTitle
   endif
   return Self
   
METHOD BarGapRatio( nBarGapRatio ) CLASS GraphPlus
   if PCount() == 0
      return ::nBarGapRatio
   else
      ::nBarGapRatio := nBarGapRatio
   endif
   return Self

METHOD BarGapWidthRatio( nBarGapWidthRatio ) CLASS GraphPlus
   if PCount() == 0
      return ::nGapWidthRatio
   else
      ::nGapWidthRatio := nBarGapWidthRatio
   endif
   return Self
   
METHOD TitleFont( aTitleFont ) CLASS GraphPlus
   if PCount() == 0
      return ::aTitleFont
   else
      ::aTitleFont := aTitleFont
   endif
   return Self

METHOD LegendFont( aLegendFont ) CLASS GraphPlus
   if PCount() == 0
      return ::aLegendFont
   else
      ::aLegendFont := aLegendFont
   endif
   return Self
   
METHOD Height( nHeight ) CLASS GraphPlus
   if PCount() == 0
      return ::nImageHeight
   else
      ::nImageHeight := nHeight
   endif
   return Self

METHOD Width( nWidth ) CLASS GraphPlus
   if PCount() == 0
      return ::nImageWidth
   else
      ::nImageWidth := nWidth
   endif
   return Self

METHOD ShowLegends( lShowLegends ) CLASS GraphPlus
   if PCount() == 0
      return ::lShowLegends
   else
      ::lShowLegends := lShowLegends
   endif
   return Self

METHOD ShowLabels( lDataLabels ) CLASS GraphPlus
   if PCount() == 0
      return ::lDataLabels
   else
      ::lDataLabels := lDataLabels
   endif
   return Self

METHOD ShowCategories( lShowCategories ) CLASS GraphPlus
   if PCount() == 0
      return ::lShowCategories
   else
      ::lShowCategories := lShowCategories
   endif
   return Self
   

METHOD ShowValues( lShowValues ) CLASS GraphPlus
   if PCount() == 0
      return ::lShowValues
   else
      ::lShowValues := lShowValues
   endif
   return Self

METHOD ShowHGrid( lHGrid ) CLASS GraphPlus
   if PCount() == 0
      return ::lHGrid
   else
      ::lHGrid := lHGrid
   endif
   return Self

METHOD ShowVGrid( lVGrid ) CLASS GraphPlus
   if PCount() == 0
      return ::lVGrid
   else
      ::lVGrid := lVGrid
   endif
   return Self
   
METHOD Picture( cPicture ) CLASS GraphPlus
   if PCount() == 0
      return ::cPicture
   else
      ::cPicture := cPicture
   endif
   return Self

METHOD PictureX( cPicture ) CLASS GraphPlus
   if PCount() == 0
      return ::cPictureX
   else
      ::cPictureX := cPicture
   endif
   return Self

METHOD PictureY( cPicture ) CLASS GraphPlus
   if PCount() == 0
      return ::cPictureY
   else
      ::cPictureY := cPicture
   endif
   return Self

METHOD ScatterLine( lScatterLine ) CLASS GraphPlus
   if PCount() == 0
      return ::lScatterLine
   else
      ::lScatterLine := lScatterLine
   endif
   return Self

METHOD VDivisions( nVDivision ) CLASS GraphPlus
   if PCount() == 0
      return ::nVDivision
   else
      ::nVDivision := nVDivision
   endif
   return Self

METHOD HDivisions( nHDivision ) CLASS GraphPlus
   if PCount() == 0
      return ::nHDivision
   else
      ::nHDivision := nHDivision
   endif
   return Self

METHOD TotalItems( aTotalItems ) CLASS GraphPlus
   if PCount() == 0
      return ::aTotalItems
   else
      ::aTotalItems := aTotalItems
   endif
   return Self

Method LegendPos( nLegendPos ) CLASS GraphPlus
   if PCount() == 0
      return ::nLegendPos
   else
      ::nLegendPos := nLegendPos
   endif
   return Self

Method ColorTheme( nColorTheme ) CLASS GraphPlus
   if PCount() == 0
      return ::nColorTheme
   else
      ::nColorTheme := nColorTheme
      asize( ::aColors, 0 ) // reset all colors if any
   endif
   return Self

   
Method Draw CLASS GraphPlus
   local aSize := {}, nLowerUpperRatio := 0
   local nDistance := 0
   local nRow := 0, nPixels := 0
   local nCount := 0
   local aCategories := {}
   local nZero := 0
   local nMaxWidth := 0, nMaxHeight := 0
   local i := 0
   LOCAL nMax, nMin, nI, nJ
   LOCAL nDivision, cDivision, nDecimalLength, nDecimal, cDecimal
   local nPositiveSum, nNegativeSum
   local aTemp := {}, aBalances := {}, nBalance := 0
   if hmg_len( ::aData ) == 0
      // nothing to draw!
      msgstop( 'No Graph data!' ) 
      return nil
   endif
   ::nTop := 10
   ::nLeft := 10
   ::nHeight := ::nImageHeight
   ::nWidth := ::nImageWidth
   ::nBottom := ::nHeight - 10
   ::nRight := ::nWidth - 10
   // init bitmap
   ::hBitmap := BT_BitmapCreateNew ( ::nWidth, ::nHeight, ::aBackColor )
   ::hDC := BT_CreateDC( ::hBitmap, BT_HDC_BITMAP, @::BTStruct )
   ::aTextColor := GetNegativeColor ( ::aBackColor )
   if hmg_len( ::aLegends ) == 0
      if ::nGraphType == GT_WATERFALL  // only three legends for waterfall
         ::aLegends := { 'Increase', 'Decrease', 'Total' }
      else
         for i := 1 to hmg_len( ::aData )
            aadd( ::aLegends, "Legend " + alltrim( str( i ) ) )
         next i
      endif   
   endif   
   if hmg_len( ::aCategories ) == 0
      // categories not defined
      for i := 1 to hmg_len( ::aData )
         aadd( ::aCategories, 'Category ' + alltrim( str( i ) ) )
      next i
   endif
   // find datalabel height
   aSize := BT_DrawTextSize (::hDC, '0', ::aLegendFont[ 1 ], ::aLegendFont[ 2 ], if( ::aLegendFont[ 3 ], BT_TEXT_BOLD, nil ))
   ::nLabelHeight := aSize[ 2 ] 
   // draw outside border
   if ::lOutBorder
      BT_DrawRectangle (::hDC, 0, 0, ::nWidth - 1 , ::nHeight - 1, ::aTextColor, 1 )
   endif   
   ::InitColors()    // init colors
   if ::nGraphType == GT_FUNNEL // Funnel Chart
      ::DrawFunnelGraph()
      BT_DeleteDC( ::BTstruct )
      return nil
   endif
   if ::nGraphType == GT_SCATTERXY
      ::DrawScatterXYGraph()
      BT_DeleteDC( ::BTstruct )
      return nil
   endif
   
   // draw legends
   if ::lShowLegends .and. hmg_len( ::aLegends ) > 0
      if .not. ( ::nGraphType == GT_SUNBURST .or. ::nGraphType == GT_TREEMAP )// don't draw now
         ::DrawLegends()
      endif   
   endif   
   // draw title
   if hmg_len( ::cTitle ) > 0 
      aSize := BT_DrawTextSize (::hDC, ::cTitle, ::aTitleFont[ 1 ], ::aTitleFont[ 2 ], if( ::aTitleFont[ 3 ], BT_TEXT_BOLD, nil ))
      BT_DrawText ( ::hDC, ::nTop / 2, ( ::nWidth - if( ::lShowLegends .AND. ::nLegendPos < 1, ::nLegendWidth, 0 ) ) / 2, ::cTitle, ::aTitleFont[ 1 ], ::aTitleFont[ 2 ], ::aTitleColor, ::aBackColor, if( ::aTitleFont[ 3 ], BT_TEXT_BOLD, nil ), BT_TEXT_CENTER )
      ::nTop := ::nTop + aSize[ 2 ]
   endif
   if .not. ( ::nGraphType == GT_PIE .or. ::nGraphType == GT_DOUGHNUT .or. ::nGraphType == GT_SUNBURST .or. ::nGraphType == GT_TREEMAP )
      ::DrawXYTitles()    // draw x/y titles
      // draw graph border
      && BT_DrawRectangle (::hDC, ::nTop, ::nLeft, ::nRight - ::nLeft , ::nBottom - ::nTop, ::aTextColor, 1 )
      
      // find max and min value
      if .not. ( ::nGraphType == GT_STACKEDCOLUMN .or. ::nGraphType == GT_STACKEDBAR )
         if ::nGraphType <> GT_WATERFALL
            nMax := 0
            nMin := 0
            FOR nJ := 1 TO HMG_LEN( ::aData )
               FOR nI :=1 TO HMG_LEN( ::aData[ nJ ] )
                  nMax := Max( ::aData[nJ][nI], nMax )
                  nMin := Min( ::aData[nJ][nI], nMin )
               NEXT nI
            NEXT nJ
         else
            aTemp := {}
            aBalances := {}
            for i := 1 to hmg_len( ::aData[ 1 ] )
               aadd( aTemp, .f. )
            next i
            for i := 1 to hmg_len( ::aTotalItems )
               aTemp[ ::aTotalItems[ i ] ] := .t.
            next i   
            // find balances
            aadd( aBalances, ::aData[ 1, 1 ] ) // first item            
            nBalance := ::aData[ 1, 1 ]
            for i := 2 to hmg_len( ::aData[ 1 ] )
               if .not. aTemp[ i ] // totalitem
                  nBalance := nBalance + ::aData[ 1, i ] // increase or decrease
               endif   
               aadd( aBalances, nBalance )
            next i
            aadd( ::aData, aclone( aTemp ) )
            aadd( ::aData, aclone( aBalances ) )
            nMax := 0
            nMin := 0
            for i := 1 to hmg_len( aBalances )
               nMax := max( nMax, aBalances[ i ] )
               nMin := min( nMin, aBalances[ i ] )
            next i            
         endif   
      else
         nMax := 0
         nMin := 0
         for nJ := 1 to hmg_len( ::aData[ 1 ] )
            nPositiveSum := 0         
            nNegativeSum := 0
            for nI := 1 to hmg_len( ::aData )
               if ::aData[ nI, nJ ] > 0
                  nPositiveSum += ::aData[ nI, nJ ]
               else
                  nNegativeSum += ::aData[ nI, nJ ]
               endif
            next nI
            nMax := max( nMax, nPositiveSum )
            nMin := min( nMin, nNegativeSum )
         next nJ
      endif
      
      if nMax > 0
         ::nUpper := nMax
      else 
         ::nUpper := 0
      endif   
      if nMin < 0
         ::nLower := nMin
      else
         ::nLower := 0   
      endif
      ::nRange := ::nUpper - ::nLower
      if ::nUpper <> 0 .and. ::nLower <> 0
         if abs( ::nLower ) < ::nUpper 
            nLowerUpperRatio := abs( ::nLower ) / ::nRange
            ::nLowerDivision := ::nVDivision * nLowerUpperRatio
            if ::nLowerDivision - int( ::nLowerDivision ) > 0.0
               ::nLowerDivision := int( ::nLowerDivision ) + 1
            endif
            ::nUpperDivision := ::nVDivision - ::nLowerDivision
            nDivision := ::nUpper / ::nUpperDivision
         else
            nLowerUpperRatio := abs( ::nUpper ) / ::nRange
            ::nUpperDivision := ::nVDivision * nLowerUpperRatio
            if ::nUpperDivision - int( ::nUpperDivision ) > 0.0
               ::nUpperDivision := int( ::nUpperDivision ) + 1
            endif   
            ::nLowerDivision := ::nVDivision - ::nUpperDivision
            nDivision := abs( ::nLower ) / ::nLowerDivision
         endif
      else
         nDivision := ::nRange / ::nVDivision
         if ::nUpper > 0
            ::nUpperDivision := ::nVDivision
            ::nLowerDivision := 0
         else
            ::nLowerDivision := ::nVDivision
            ::nUpperDivision := 0
         endif         
      endif         
      if int( nDivision ) > 0
         cDivision := alltrim( str( int( nDivision ), 14, 0 ) )
         if hmg_len( cDivision ) > 1 // if more than 1 digits round to power of 10s
            nDivision := ( val( left( cDivision, 2 ) ) + 1 ) * 10^ ( hmg_len( cDivision ) - 2 )
         else
            if nDivision > int( nDivision ) 
               nDivision++
            endif
         endif
      else //
         cDivision := transform( nDivision, '9.9999999' )
         nDecimalLength := 8 - hmg_len( alltrim( str( val( right( cDivision, 7 ) ) ) ) )
         nDecimal := val( right( cDivision, 7 ) )
         cDecimal := alltrim( str( nDecimal ) )
         nDivision := (val( alltrim( left( cDecimal, 1 ) )) + 1 ) /(10 ^nDecimalLength)
      endif
      ::nRange := ::nVDivision * nDivision 
      if hmg_len( ::cPicture ) == 0
         ::cPicture := replicate( '9', max( hmg_len( alltrim( str( ::nUpper ) ) ), hmg_len( alltrim( str( ::nLower ) ) ) ) )
      endif   
      // find scales
      asize( ::aScale, 0 )
      nDistance := ::nRange / ::nVDivision
      for i := ::nUpperDivision to 1 step -1
         aadd( ::aScale, alltrim( transform( nDistance * i, ::cPicture ) ) )
      next i
      aadd( ::aScale, alltrim( transform( 0, ::cPicture ) ) )
      nZero := hmg_len( ::aScale )
      for i := 1 to ::nLowerDivision
         aadd( ::aScale, '-' + alltrim( transform( nDistance * i, ::cPicture ) ) )
      next i
      // find maxscaleitemwidth
      nMaxWidth := 0
      for i := 1 to hmg_len( ::aScale )
         aSize := BT_DrawTextSize (::hDC, ::aScale[ i ], ::aLegendFont[ 1 ], ::aLegendFont[ 2 ], if( ::aLegendFont[ 3 ], BT_TEXT_BOLD, nil ))
         nMaxWidth := max( nMaxWidth, aSize[ 1 ] )
      next i
      ::nMaxScaleWidth := nMaxWidth
      // draw all items other than graph
      ::nTop := ::nTop + 20
      ::nBottom := ::nBottom - 20
      if ::lShowCategories 
         ::DrawCategories()
      endif
      // scale
      if ::lShowValues 
         if .not. ::lGraphRotated // not bar
            ::DrawVerticalScale()
         else
            ::DrawBarScale()
         endif   
      endif
      ::nRight := ::nRight - 5
      ::nLeft := ::nLeft + 5
      // find zero position
      do case
      case nZero == 1 // all are negative!
         if .not. ::lGraphRotated //bar
            ::nZeroPos := ::nTop
         else
            ::nZeroPos := ::nRight
         endif
      case nZero == hmg_len( ::aScale ) // all are positive!
         if .not. ::lGraphRotated //other than bar
            ::nZeroPos := ::nBottom
         else
            ::nZeroPos := ::nLeft
         endif
      otherwise  // zero position in the middles
         if .not. ::lGraphRotated // other than bar
            ::nZeroPos := ::nTop + ( ( ::nBottom - ::nTop ) / ::nVDivision * ( nZero - 1 ) )
            ::nPixelperPoint := ( ::nBottom - ::nTop ) / ::nRange
         else // bar
            ::nZeroPos := ::nRight - ( ( ::nRight - ::nLeft ) / ::nVDivision * ( nZero - 1 ) )
            ::nPixelperPoint := ( ::nRight - ::nLeft ) / ::nRange
         endif
      endcase
      ::nZeroPos := int( ::nZeroPos )
      if ::lHGrid 
         ::DrawHorizontalGrid()
      endif
      if ::lVGrid 
         ::DrawVerticalGrid()
      endif
   endif   
   do case
   case ::nGraphType == GT_COLUMNS
      ::DrawColumnGraph()
   case ::nGraphType == GT_LINE
      ::DrawLineGraph()
   case ::nGraphType == GT_POINTS
      ::DrawPointsGraph()
   case ::nGraphType == GT_PIE
      ::DrawPieGraph()
   case ::nGraphType == GT_BAR
      ::DrawBarGraph()
   case ::nGraphType == GT_STACKEDCOLUMN
      ::DrawStackedColumnGraph()   
   case ::nGraphType == GT_STACKEDBAR
      ::DrawStackedBarGraph()   
   case ::nGraphType == GT_AREA
      ::DrawAreaGraph()   
   case ::nGraphType == GT_DOUGHNUT 
      ::DrawDoughnutGraph()   
   case ::nGraphType == GT_SUNBURST 
      ::DrawSunBurstGraph()   
   case ::nGraphType == GT_WATERFALL
      ::DrawWaterfallGraph()
   case ::nGraphType == GT_TREEMAP
      ::DrawTreeMapGraph()   
   endcase
   BT_DeleteDC( ::BTstruct )
return nil   

Method DrawColumnGraph Class GraphPlus
   local nCategories := 0
   local nRow := 0, nCol := 0, nData := 0
   local nTotalBars := 0
   local nAvailableWidth := 0
   local nTotalBarRatio := 0
   local nCategoryRatio := 0
   local nTotalBarSize := 0

   local nBarWidth := 0
   local nCategoryWidth := 0
   local nFactor := 0
   local nBarGap := 0
   
   
   if hmg_len( ::aData ) > 0
      if valtype( ::aData[ 1 ] ) == 'A'
         nCategories := hmg_len( ::aData[ 1 ] )
      endif
   endif
   if nCategories == 0
      return nil
   endif
   nTotalBars := nCategories * hmg_len( ::aData )
   nAvailableWidth := ::nRight - ::nLeft
   nTotalBarRatio := nTotalBars * 1 // 100%
   nTotalBarRatio := nTotalBarRatio + ( nTotalBarRatio - nCategories ) * ::nBarGapRatio 
   nCategoryRatio := nCategories * ::nGapWidthRatio // n Times of Bar size for each category
   nTotalBarRatio := nTotalBarRatio + nCategoryRatio

   nBarWidth := nAvailableWidth / nTotalBarRatio // 100% width
   nCategoryWidth := nBarWidth * ::nGapWidthRatio
   nBarGap      := nBarWidth * ::nBarGapRatio
   
   nCol := ::nLeft + ( nCategoryWidth / 2 )
   nRow := ::nZeroPos
   for i := 1 to hmg_len( ::aData[ 1 ] )
      for j := 1 to hmg_len( ::aData )
         nData := ::aData[ j, i ]
         nBarHeight := int( ( ::nBottom - ::nTop ) * ( abs( nData ) / ::nRange ) )
         if nData > 0
            // positive
            BT_DrawFillRectangle ( ::hDC, nRow - nBarHeight, nCol, nBarWidth, nBarHeight, ::aColors[ j ], ::aColors[ j ], 1 )
            if ::lDataLabels
               BT_DrawText ( ::hDC, nRow - nBarHeight - 15, nCol + (nBarWidth / 2 ), alltrim( transform( nData, ::cPicture ) ), ::aLegendFont[ 1 ], ::aLegendFont[ 2 ], ::aTextColor, ::aBackColor, if( ::aLegendFont[ 3 ], BT_TEXT_BOLD, nil ), BT_TEXT_CENTER )
            endif
         else
            // negative
            BT_DrawFillRectangle ( ::hDC, nRow, nCol, nBarWidth, nBarHeight, ::aColors[ j ], ::aColors[ j ], 1 )
            if ::lDataLabels
               BT_DrawText ( ::hDC, nRow + nBarHeight + 3, nCol + (nBarWidth / 2 ), if( nData < 0, '-', '' ) + alltrim( transform( abs( nData ), ::cPicture ) ), ::aLegendFont[ 1 ], ::aLegendFont[ 2 ], ::aTextColor, ::aBackColor, if( ::aLegendFont[ 3 ], BT_TEXT_BOLD, nil ), BT_TEXT_CENTER )
            endif
         endif
         nCol := nCol + nBarWidth 
         if j < hmg_len( ::aData )
            nCol := nCol + nBarGap
         endif   
      next j
      nCol := nCol + nCategoryWidth
   next i
return nil

Method DrawVerticalScale Class GraphPlus
   local nScaleItems := 0
   local nRow := 0, nCol := 0, nData := 0
   local nAvailableHeight := 0
   local nTotalBarRatio := 0
   local nBarHeight := 0
   local nScaleItemHeight := 0
   local aSize := {}, i := 0
   local nMaxWidth := 0
   nScaleItems := hmg_len( ::aScale )
   if nScaleItems == 0
      return nil
   endif
   nAvailableHeight := ::nBottom - ::nTop
   nTotalBarRatio := nScaleItems - 1 // equal space between scales
   nScaleItemHeight := nAvailableHeight / nTotalBarRatio 
   nRow := ::nTop
   for i := 1 to hmg_len( ::aScale )
      aSize := BT_DrawTextSize (::hDC, ::aScale[ i ], ::aLegendFont[ 1 ], ::aLegendFont[ 2 ], if( ::aLegendFont[ 3 ], BT_TEXT_BOLD, nil ))
      nMaxWidth := max( nMaxWidth, aSize[ 1 ] )
   next i  
   ::nLeft := ::nLeft + nMaxWidth
   for i := 1 to hmg_len( ::aScale )
      BT_DrawText ( ::hDC, nRow + 2, ::nLeft, ::aScale[ i ], ::aLegendFont[ 1 ], ::aLegendFont[ 2 ], ::aTextColor, ::aBackColor, if( ::aLegendFont[ 3 ], BT_TEXT_BOLD, nil ),  BT_TEXT_BASELINE + BT_TEXT_RIGHT)
      nRow := nRow + nScaleItemHeight
   next i
return nil

Method DrawHorizontalGrid Class GraphPlus
   local i := 0
   local nHeight := ( ::nBottom - ::nTop ) / ::nVDivision
   local nRow := 0
   nRow := ::nTop
   if .not. ::lGraphRotated // not bar
      for i := 1 to hmg_len( ::aScale )
         BT_DrawLine (::hDC, nRow, ::nLeft, nRow, ::nRight, ::aGridLineColor, ::nLineWidth)
         nRow := nRow + nHeight
      next i
   else 
      nHeight := ( ::nBottom - ::nTop ) / hmg_len( ::aCategories )
      nRow := ::nTop
      for i := 1 to hmg_len( ::aCategories )
         BT_DrawLine (::hDC, nRow, ::nLeft, nRow, ::nRight, ::aGridLineColor, ::nLineWidth)
         nRow := nRow + nHeight
      next i
      BT_DrawLine (::hDC, nRow, ::nLeft, nRow, ::nRight, ::aGridLineColor, ::nLineWidth)
   endif   
return nil

Method DrawVerticalGrid Class GraphPlus
   local nCategories := 0
   local nRow := 0, nCol := 0, nData := 0, nWidth := 0
   local nAvailableWidth := 0
   local nTotalBarRatio := 0
   local nBarWidth := 0
   local nCategoryWidth := 0
   if .not. ::lGraphRotated // columns and like
      nCategories := hmg_len( ::aData[ 1 ] )
      if nCategories == 0
         return nil
      endif
      nAvailableWidth := ::nRight - ::nLeft
      if ::nGraphType <> GT_AREA // Area charts starts at left and ends at right
         nTotalBarRatio := nCategories // equal space between categories
         nCategoryWidth := nAvailableWidth / nTotalBarRatio // 100% width
      else   
         nTotalBarRatio := nCategories - 1 // equal space between categories
         nCategoryWidth := nAvailableWidth / nTotalBarRatio // 100% width
      endif   
      BT_DrawLine (::hDC, ::nTop, ::nLeft, ::nBottom, ::nLeft, ::aGridLineColor, ::nLineWidth)
      nCol := ::nLeft
      for i := 1 to nCategories
         nCol := nCol + nCategoryWidth
         BT_DrawLine (::hDC, ::nTop, nCol, ::nBottom, nCol, ::aGridLineColor, ::nLineWidth)
      next i
   else
      nWidth := ( ::nRight - ::nLeft ) / ( hmg_len( ::aScale ) - 1 )
      nCol := ::nLeft
      for i := 1 to hmg_len( ::aScale )
         BT_DrawLine (::hDC, ::nTop, nCol, ::nBottom, nCol, ::aGridLineColor, ::nLineWidth)
         nCol := nCol + nWidth
      next i
      BT_DrawLine (::hDC, ::nTop, nCol, ::nBottom, nCol, ::aGridLineColor, ::nLineWidth)
   endif   
return nil

Method DrawBarScale Class GraphPlus
   local nScaleItems := 0
   local nRow := 0, nCol := 0, nData := 0
   local nAvailableWidth := 0
   local nTotalBarRatio := 0
   local nBarWidth := 0
   local nCategoryWidth := 0
   nScaleItems := hmg_len( ::aScale ) 
   if nScaleItems == 0
      return nil
   endif
   nAvailableWidth := ::nRight - ::nLeft
   nTotalBarRatio := nScaleItems - 1 // equal space between categories
   nScaleItemWidth := nAvailableWidth / nTotalBarRatio // 100% width
   nCol := ::nLeft
   for i := hmg_len( ::aScale ) to 1 step -1
      BT_DrawText ( ::hDC, ::nBottom, nCol, ::aScale[ i ], ::aLegendFont[ 1 ], ::aLegendFont[ 2 ], ::aTextColor, ::aBackColor, if( ::aLegendFont[ 3 ], BT_TEXT_BOLD, nil ),  BT_TEXT_CENTER)
      nCol := nCol + nScaleItemWidth
   next i 
return nil


Method DrawLineGraph Class GraphPlus
   local nCategories := 0
   local nRow := 0, nCol := 0, nData := 0, nFirstCol := 0, nPrevRow := 0, nPrevCol := 0
   local nTotalBars := 0
   local nAvailableWidth := 0
   local nTotalBarRatio := 0
   local nCategoryRatio := 0
   local nTotalBarSize := 0

   local nBarWidth := 0
   local nCategoryWidth := 0
   local nFactor := 0
   local nBarGap := 0
   
   if hmg_len( ::aData ) > 0
      if valtype( ::aData[ 1 ] ) == 'A'
         nCategories := hmg_len( ::aData[ 1 ] )
      endif
   endif
   if nCategories == 0
      return nil
   endif
   nAvailableWidth := ::nRight - ::nLeft
   nTotalBarRatio := nCategories // equal space between categories

   nBarWidth := nAvailableWidth / nTotalBarRatio // 100% width
   nCategoryWidth := nBarWidth 
   nFirstCol := ::nLeft + ( nCategoryWidth / 2 )   
   nCol := ::nLeft + ( nCategoryWidth / 2 )
   nRow := ::nZeroPos
   for j := 1 to hmg_len( ::aData )
      nCol := nFirstCol
      for i := 1 to hmg_len( ::aData[ 1 ] )
         nData := ::aData[ j, i ]
         nBarHeight := int( ( ::nBottom - ::nTop ) * ( abs( nData ) / ::nRange ) )
         if nData > 0
            // positive
            BT_DrawFillEllipse ( ::hDC, nRow - nBarHeight - ::nPointSize/2, nCol - ::nPointSize / 2, ::nPointSize, ::nPointSize, ::aColors[ j ], ::aColors[ j ], 1)
            if ::lDataLabels
               BT_DrawText ( ::hDC, nRow - nBarHeight - 15 - ::nPointSize / 2 , nCol, alltrim( transform( nData, ::cPicture ) ), ::aLegendFont[ 1 ], ::aLegendFont[ 2 ], ::aTextColor, ::aBackColor, if( ::aLegendFont[ 3 ], BT_TEXT_BOLD, nil ), BT_TEXT_CENTER )
            endif
            if i <> 1 // first point
               BT_DrawLine (::hDC, nPrevRow, nPrevCol, nRow - nBarHeight, nCol, ::aColors[ j ], ::nLineWidth)              
            endif
            nPrevRow := nRow - nBarHeight
            nPrevCol := nCol
         else
            // negative
            BT_DrawFillEllipse ( ::hDC, nRow + nBarHeight - ::nPointSize/2, nCol - ::nPointSize/2, ::nPointSize, ::nPointSize, ::aColors[ j ], ::aColors[ j ], 1 )
            if ::lDataLabels
               BT_DrawText ( ::hDC, nRow + nBarHeight + 3 + ::nPointSize / 2 , nCol, if( nData < 0, '-', '' ) + alltrim( transform( abs( nData ), ::cPicture ) ), ::aLegendFont[ 1 ], ::aLegendFont[ 2 ], ::aTextColor, ::aBackColor, if( ::aLegendFont[ 3 ], BT_TEXT_BOLD, nil ), BT_TEXT_CENTER )
            endif
            if i <> 1
               BT_DrawLine (::hDC, nPrevRow, nPrevCol, nRow + nBarHeight, nCol, ::aColors[ j ], ::nLineWidth)              
            endif
            nPrevRow := nRow + nBarHeight
            nPrevCol := nCol
         endif
         nCol := nCol + nCategoryWidth
      next i
   next j

return nil

Method DrawPointsGraph Class GraphPlus
   local nCategories := 0
   local nRow := 0, nCol := 0, nData := 0
   local nTotalBars := 0
   local nAvailableWidth := 0
   local nTotalBarRatio := 0
   local nCategoryRatio := 0
   local nTotalBarSize := 0

   local nBarWidth := 0
   local nCategoryWidth := 0
   local nFactor := 0
   local nBarGap := 0
   
   if hmg_len( ::aData ) > 0
      if valtype( ::aData[ 1 ] ) == 'A'
         nCategories := hmg_len( ::aData[ 1 ] )
      endif
   endif
   if nCategories == 0
      return nil
   endif
   nAvailableWidth := ::nRight - ::nLeft
   nTotalBarRatio := nCategories // equal space between categories

   nBarWidth := nAvailableWidth / nTotalBarRatio // 100% width
   nCategoryWidth := nBarWidth 
   
   nCol := ::nLeft + ( nCategoryWidth / 2 )
   nRow := ::nZeroPos
   for i := 1 to hmg_len( ::aData[ 1 ] )
      for j := 1 to hmg_len( ::aData )
         nData := ::aData[ j, i ]
         nBarHeight := int( ( ::nBottom - ::nTop ) * ( abs( nData ) / ::nRange ) )
         if nData > 0
            // positive
            BT_DrawFillEllipse ( ::hDC, nRow - nBarHeight - ::nPointSize / 2, nCol - ::nPointSize / 2, ::nPointSize, ::nPointSize, ::aColors[ j ], ::aColors[ j ], 1)
            if ::lDataLabels
               BT_DrawText ( ::hDC, nRow - nBarHeight - 15 - ::nPointSize / 2 , nCol, alltrim( transform( nData, ::cPicture ) ), ::aLegendFont[ 1 ], ::aLegendFont[ 2 ], ::aTextColor, ::aBackColor, if( ::aLegendFont[ 3 ], BT_TEXT_BOLD, nil ), BT_TEXT_CENTER )
            endif
         else
            // negative
            BT_DrawFillEllipse ( ::hDC, nRow + nBarHeight - ::nPointSize / 2, nCol - ::nPointSize / 2, ::nPointSize, ::nPointSize, ::aColors[ j ], ::aColors[ j ], 1 )
            if ::lDataLabels
               BT_DrawText ( ::hDC, nRow + nBarHeight + 3 + ::nPointSize / 2 , nCol, if( nData < 0, '-', '' ) + alltrim( transform( abs( nData ), ::cPicture ) ), ::aLegendFont[ 1 ], ::aLegendFont[ 2 ], ::aTextColor, ::aBackColor, if( ::aLegendFont[ 3 ], BT_TEXT_BOLD, nil ), BT_TEXT_CENTER )
            endif
         endif
      next j
      nCol := nCol + nCategoryWidth
   next i
return nil

Method DrawPieGraph Class GraphPlus
   local aPoints := {}
   LOCAL aDegrees := {}
   LOCAL aCumulative := {}
   LOCAL j,i,sum := 0, lsum := 0
   LOCAL shadowcolor
   LOCAL previous_cumulative
   local aSerieValues := {}
   local nChartSize := 0
   local aLabelCumulative := {}
   local ser_sum := 0
   
   for i := 1 to hmg_len( ::aData[ 1 ] )
      aadd( aSerieValues, ::aData[ 1, i ] )
   next i   
   ::nLeft += 25
   ::nRight -= 25
   ::nBottom -= 25
   ::nTop += 25
   nChartSize :=  min( ::nRight - ::nLeft, ::nBottom - ::nTop ) 
   aPoints := ::FindPieCoordinates( nChartSize )
   for i := 1 to HMG_LEN(aSerieValues)
      ser_sum += aSerieValues[i]
   next i
   for i := 1 to HMG_LEN(aSerieValues)
      aadd(aDegrees,Round(aSerieValues[i]/ser_sum * 360,0))
   next i
   sum := 0
   for i := 1 to HMG_LEN(aDegrees)
      sum += aDegrees[i]
   next i
   if sum <> 360
      aDegrees[HMG_LEN(aDegrees)] := aDegrees[HMG_LEN(aDegrees)] + (360 - sum)
   endif

   sum := 0
   lSum := 0
   for i := 1 to HMG_LEN(aDegrees)
      sum := sum + aDegrees[i]
      aadd(aCumulative,sum)
      lSum := lSum + ( aDegrees[i] / 2 )
      aadd( aLabelCumulative, lSum )
      lSum := lSum + ( aDegrees[i] / 2 )
   next i
   for i := 1 to hmg_len( aCumulative )
      if aCumulative[ i ] > 90
         aCumulative[ i ] := 450 - aCumulative[ i ]
      else
         aCumulative[ i ] := 90 - aCumulative[ i ]
      endif
      if aLabelCumulative[ i ] > 90
         aLabelCumulative[ i ] := 450 - aLabelCumulative[ i ]
      else
         aLabelCumulative[ i ] := 90 - aLabelCumulative[ i ]
      endif
   next i   
      

   previous_cumulative := -1

   nFromRadialRow := aPoints[ 3 ] // toprightrow
   nFromRadialCol := aPoints[ 9 ] // middletopcol
   for i := 1 to HMG_LEN(aCumulative)
      if aCumulative[i] == previous_cumulative
         loop
      endif
      previous_cumulative := aCumulative[i]
      do case
      case aCumulative[i] <= 45
         nToRadialCol := aPoints[ 14 ] // middlerightcol
         nToRadialRow := aPoints[ 13 ] - Round( aCumulative[i] / 45 * ( aPoints[ 13 ] - aPoints[ 3 ] ),0)
      case aCumulative[i] <= 90 .and. aCumulative[i] > 45
         nToRadialRow := aPoints[ 3 ]
         nToRadialCol := aPoints[ 4 ] - Round((aCumulative[i] - 45) / 45 * ( aPoints[ 4 ] - aPoints[ 9 ]),0)
      case aCumulative[i] <= 135 .and. aCumulative[i] > 90
         nToRadialRow := aPoints[ 1 ]
         nToRadialCol := aPoints[ 9 ] - Round((aCumulative[i] - 90) / 45 * ( aPoints[ 9 ] - aPoints[ 2 ]),0)
      case aCumulative[i] <= 180 .and. aCumulative[i] > 135
         nToRadialCol := aPoints[ 2 ]
         nToRadialRow := aPoints[ 1 ] + Round((aCumulative[i] - 135) / 45 * ( aPoints[ 10 ] - aPoints[ 1 ]),0)
      case aCumulative[i] <= 225 .and. aCumulative[i] > 180
         nToRadialCol := aPoints[ 2 ]
         nToRadialRow := aPoints[ 10 ] + Round((aCumulative[i] - 180) / 45 * ( aPoints[ 7 ] - aPoints[ 10 ]),0)
      case aCumulative[i] <= 270 .and. aCumulative[i] > 225
         nToRadialRow := aPoints[ 7 ]
         nToRadialCol := aPoints[ 8 ] + Round((aCumulative[i] - 225) / 45 * ( aPoints[ 12 ] - aPoints[ 8 ]),0)
      case aCumulative[i] <= 315 .and. aCumulative[i] > 270
         nToRadialRow := aPoints[ 7 ]
         nToRadialCol := aPoints[ 12 ] + Round((aCumulative[i] - 270) / 45 * ( aPoints[ 6 ] - aPoints[ 12 ]),0)
      case aCumulative[i] <= 360 .and. aCumulative[i] > 315
         nToRadialCol := aPoints[ 6 ]
         nToRadialRow := aPoints[ 5 ] - Round((aCumulative[i] - 315) / 45 * (aPoints[ 5 ] - aPoints[ 13 ]),0)
      endcase
      BT_DrawPie ( ::hDC, aPoints[ 1 ], aPoints[ 2 ], aPoints[ 5 ], aPoints[ 4 ], nToRadialRow, nToRadialCol, nFromRadialRow, nFromRadialCol, ::aBackColor, ::nPieGap, ::aColors[ i ] )
      nFromRadialRow := nToRadialRow
      nFromRadialCol := nToRadialCol
   next i
   if ::lDataLabels
      nChartSize :=  nChartSize / 1.5 
      aPoints := ::FindPieCoordinates( nChartSize )
      previous_cumulative := -1
      nFromRadialRow := aPoints[ 3 ] // toprightrow
      nFromRadialCol := aPoints[ 9 ] // middletopcol
      for i := 1 to HMG_LEN(aLabelCumulative)
         if aLabelCumulative[i] == previous_cumulative
            loop
         endif
         cValue := AllTrim( Transform( aSerieValues[ i ], ::cPicture ) ) + ' (' + AllTrim( Transform( aSerieValues[ i ] / ser_sum * 100, '999.99' ) ) + '%)'
         previous_cumulative := aLabelCumulative[i]
         do case
         case aLabelCumulative[i] <= 45
            nToRadialCol := aPoints[ 14 ] // middlerightcol
            nToRadialRow := aPoints[ 13 ] - Round( aLabelCumulative[i] / 45 * ( aPoints[ 13 ] - aPoints[ 3 ] ),0)
         case aLabelCumulative[i] <= 90 .and. aLabelCumulative[i] > 45
            nToRadialRow := aPoints[ 3 ]
            nToRadialCol := aPoints[ 4 ] - Round((aLabelCumulative[i] - 45) / 45 * ( aPoints[ 4 ] - aPoints[ 9 ]),0)
         case aLabelCumulative[i] <= 135 .and. aLabelCumulative[i] > 90
            nToRadialRow := aPoints[ 1 ]
            nToRadialCol := aPoints[ 9 ] - Round((aLabelCumulative[i] - 90) / 45 * ( aPoints[ 9 ] - aPoints[ 2 ]),0)
         case aLabelCumulative[i] <= 180 .and. aLabelCumulative[i] > 135
            nToRadialCol := aPoints[ 2 ]
            nToRadialRow := aPoints[ 1 ] + Round((aLabelCumulative[i] - 135) / 45 * ( aPoints[ 10 ] - aPoints[ 1 ]),0)
         case aLabelCumulative[i] <= 225 .and. aLabelCumulative[i] > 180
            nToRadialCol := aPoints[ 2 ]
            nToRadialRow := aPoints[ 10 ] + Round((aLabelCumulative[i] - 180) / 45 * ( aPoints[ 7 ] - aPoints[ 10 ]),0)
         case aLabelCumulative[i] <= 270 .and. aLabelCumulative[i] > 225
            nToRadialRow := aPoints[ 7 ]
            nToRadialCol := aPoints[ 8 ] + Round((aLabelCumulative[i] - 225) / 45 * ( aPoints[ 12 ] - aPoints[ 8 ]),0)
         case aLabelCumulative[i] <= 315 .and. aLabelCumulative[i] > 270
            nToRadialRow := aPoints[ 7 ]
            nToRadialCol := aPoints[ 12 ] + Round((aLabelCumulative[i] - 270) / 45 * ( aPoints[ 6 ] - aPoints[ 12 ]),0)
         case aLabelCumulative[i] <= 360 .and. aLabelCumulative[i] > 315
            nToRadialCol := aPoints[ 6 ]
            nToRadialRow := aPoints[ 5 ] - Round((aLabelCumulative[i] - 315) / 45 * (aPoints[ 5 ] - aPoints[ 13 ]),0)
         endcase
         BT_DrawText ( ::hDC, nToRadialRow, nToRadialCol, cValue, ::aLegendFont[ 1 ], ::aLegendFont[ 2 ], ::aTextColor, ::aColors[ i ], if( ::aLegendFont[ 3 ], BT_TEXT_BOLD, nil ), BT_TEXT_CENTER)
         nFromRadialRow := nToRadialRow
         nFromRadialCol := nToRadialCol
      next i
   endif
   
return nil

Method DrawDoughnutGraph Class GraphPlus
   LOCAL nFromRadialRow, nFromRadialCol, nToRadialRow, nToRadialCol
   LOCAL aDegrees := {}
   LOCAL aCumulative := {}
   LOCAL j,i,sum := 0, lsum := 0
   LOCAL shadowcolor
   LOCAL previous_cumulative
   local aSerieValues := {}
   local nChartSize := 0
   local aLabelCumulative := {}
   local ser_sum := 0
   local nCircles := hmg_len( ::aCategories )
   local nCircleRatio := nCircles + ( nCircles * ::nBarGapRatio )
   local nPieWidth := 0
   local nHoleSize := 0
   local nCenterRow := 0
   local nCenterCol := 0
   local aPoints := {}
   local nOriginalSize := 0
   local aColor := {}, aHSL := {}
   
   ::nLeft += 25
   ::nRight -= 25
   ::nBottom -= 25
   ::nTop += 25
   nChartSize :=  min( ::nRight - ::nLeft, ::nBottom - ::nTop ) 
   nOriginalSize := nChartSize
   nPieWidth := nChartSize / nCircleRatio
   nHoleSize := nPieWidth * ::nBarGapRatio
   if hmg_len( ::aData ) == 0
      return nil
   endif
   if .not. ( valtype( ::aData ) == 'A' .and. hmg_len( ::aData[ 1 ] ) > 0 )
      return nil
   endif
   for j := 1 to hmg_len( ::aData[ 1 ] )
      asize( aSerieValues, 0 )
      asize( aDegrees, 0 )
      asize( aCumulative, 0 )
      asize( aLabelCumulative, 0 )
      for i := 1 to hmg_len( ::aData )
         aadd( aSerieValues, ::aData[ i, j ] )
      next i  
      ser_sum := 0
      for i := 1 to HMG_LEN(aSerieValues)
         ser_sum += aSerieValues[i]
      next i
      for i := 1 to HMG_LEN(aSerieValues)
         aadd(aDegrees,Round(aSerieValues[i]/ser_sum * 360,0))
      next i
      sum := 0
      for i := 1 to HMG_LEN(aDegrees)
         sum += aDegrees[i]
      next i
      if sum <> 360
         aDegrees[HMG_LEN(aDegrees)] := aDegrees[HMG_LEN(aDegrees)] + (360 - sum)
      endif

      sum := 0
      lSum := 0
      for i := 1 to HMG_LEN(aDegrees)
         sum := sum + aDegrees[i]
         aadd(aCumulative,sum)
         lSum := lSum + ( aDegrees[i] / 2 )
         aadd( aLabelCumulative, lSum )
         lSum := lSum + ( aDegrees[i] / 2 )
      next i
      for i := 1 to hmg_len( aCumulative )
         if aCumulative[ i ] > 90
            aCumulative[ i ] := 450 - aCumulative[ i ]
         else
            aCumulative[ i ] := 90 - aCumulative[ i ]
         endif
         if aLabelCumulative[ i ] > 90
            aLabelCumulative[ i ] := 450 - aLabelCumulative[ i ]
         else
            aLabelCumulative[ i ] := 90 - aLabelCumulative[ i ]
         endif
      next i   
      previous_cumulative := -1
      aPoints := ::FindPieCoordinates( nChartSize )
      nFromRadialRow := aPoints[ 3 ] // toprightrow
      nFromRadialCol := aPoints[ 9 ] // middletopcol
      for i := 1 to HMG_LEN(aCumulative)
         if aCumulative[i] == previous_cumulative
            loop
         endif
         previous_cumulative := aCumulative[i]
         do case
         case aCumulative[i] <= 45
            nToRadialCol := aPoints[ 14 ] // middlerightcol
            nToRadialRow := aPoints[ 13 ] - Round( aCumulative[i] / 45 * ( aPoints[ 13 ] - aPoints[ 3 ] ),0)
         case aCumulative[i] <= 90 .and. aCumulative[i] > 45
            nToRadialRow := aPoints[ 3 ]
            nToRadialCol := aPoints[ 4 ] - Round((aCumulative[i] - 45) / 45 * ( aPoints[ 4 ] - aPoints[ 9 ]),0)
         case aCumulative[i] <= 135 .and. aCumulative[i] > 90
            nToRadialRow := aPoints[ 1 ]
            nToRadialCol := aPoints[ 9 ] - Round((aCumulative[i] - 90) / 45 * ( aPoints[ 9 ] - aPoints[ 2 ]),0)
         case aCumulative[i] <= 180 .and. aCumulative[i] > 135
            nToRadialCol := aPoints[ 2 ]
            nToRadialRow := aPoints[ 1 ] + Round((aCumulative[i] - 135) / 45 * ( aPoints[ 10 ] - aPoints[ 1 ]),0)
         case aCumulative[i] <= 225 .and. aCumulative[i] > 180
            nToRadialCol := aPoints[ 2 ]
            nToRadialRow := aPoints[ 10 ] + Round((aCumulative[i] - 180) / 45 * ( aPoints[ 7 ] - aPoints[ 10 ]),0)
         case aCumulative[i] <= 270 .and. aCumulative[i] > 225
            nToRadialRow := aPoints[ 7 ]
            nToRadialCol := aPoints[ 8 ] + Round((aCumulative[i] - 225) / 45 * ( aPoints[ 12 ] - aPoints[ 8 ]),0)
         case aCumulative[i] <= 315 .and. aCumulative[i] > 270
            nToRadialRow := aPoints[ 7 ]
            nToRadialCol := aPoints[ 12 ] + Round((aCumulative[i] - 270) / 45 * ( aPoints[ 6 ] - aPoints[ 12 ]),0)
         case aCumulative[i] <= 360 .and. aCumulative[i] > 315
            nToRadialCol := aPoints[ 6 ]
            nToRadialRow := aPoints[ 5 ] - Round((aCumulative[i] - 315) / 45 * (aPoints[ 5 ] - aPoints[ 13 ]),0)
         endcase
         // draw pie
         BT_DrawPie ( ::hDC, aPoints[ 1 ], aPoints[ 2 ], aPoints[ 5 ], aPoints[ 4 ], nToRadialRow, nToRadialCol, nFromRadialRow, nFromRadialCol, ::aBackColor, ::nPieGap, ::aColors[ i ] )
         nFromRadialRow := nToRadialRow
         nFromRadialCol := nToRadialCol
      next i
      nChartSize := nChartSize - nPieWidth
      aPoints := ::FindPieCoordinates( nChartSize )
      // draw the hole!
      nCenterRow := aPoints[ 10 ] - nChartSize / 2 
      nCenterCol := aPoints[ 9 ] - nChartSize / 2 
      BT_DrawFillEllipse ( ::hDC, nCenterRow, nCenterCol, nChartSize, nChartSize, ::aBackColor, ::aBackColor, 1 )
      
      nChartSize := nChartSize - nHoleSize
      aPoints := ::FindPieCoordinates( nChartSize )
   next j   
   if ::lDataLabels
      nChartSize := nOriginalSize
      for j := 1 to hmg_len( ::aData[ 1 ] )
         asize( aSerieValues, 0 )
         asize( aDegrees, 0 )
         asize( aCumulative, 0 )
         asize( aLabelCumulative, 0 )
         for i := 1 to hmg_len( ::aData )
            aadd( aSerieValues, ::aData[ i, j ] )
         next i  
         ser_sum := 0
         for i := 1 to HMG_LEN(aSerieValues)
            ser_sum += aSerieValues[i]
         next i
         for i := 1 to HMG_LEN(aSerieValues)
            aadd(aDegrees,Round(aSerieValues[i]/ser_sum * 360,0))
         next i
         sum := 0
         for i := 1 to HMG_LEN(aDegrees)
            sum += aDegrees[i]
         next i
         if sum <> 360
            aDegrees[HMG_LEN(aDegrees)] := aDegrees[HMG_LEN(aDegrees)] + (360 - sum)
         endif

         sum := 0
         lSum := 0
         for i := 1 to HMG_LEN(aDegrees)
            sum := sum + aDegrees[i]
            aadd(aCumulative,sum)
            lSum := lSum + ( aDegrees[i] / 2 )
            aadd( aLabelCumulative, lSum )
            lSum := lSum + ( aDegrees[i] / 2 )
         next i
         for i := 1 to hmg_len( aCumulative )
            if aCumulative[ i ] > 90
               aCumulative[ i ] := 450 - aCumulative[ i ]
            else
               aCumulative[ i ] := 90 - aCumulative[ i ]
            endif
            if aLabelCumulative[ i ] > 90
               aLabelCumulative[ i ] := 450 - aLabelCumulative[ i ]
            else
               aLabelCumulative[ i ] := 90 - aLabelCumulative[ i ]
            endif
         next i   
         aPoints := ::FindPieCoordinates( nChartSize - nPieWidth / 2 )
         previous_cumulative := -1
         nFromRadialRow := aPoints[ 3 ] // toprightrow
         nFromRadialCol := aPoints[ 9 ] // middletopcol
         for i := 1 to HMG_LEN(aCumulative)
            if aLabelCumulative[i] == previous_cumulative
               loop
            endif
            cValue := AllTrim( Transform( aSerieValues[ i ], ::cPicture ) ) + ' (' + AllTrim( Transform( aSerieValues[ i ] / ser_sum * 100, '999.99' ) ) + '%)'
            previous_cumulative := aLabelCumulative[i]
            do case
            case aLabelCumulative[i] <= 45
               nToRadialCol := aPoints[ 14 ] // middlerightcol
               nToRadialRow := aPoints[ 13 ] - Round( aLabelCumulative[i] / 45 * ( aPoints[ 13 ] - aPoints[ 3 ] ),0)
            case aLabelCumulative[i] <= 90 .and. aLabelCumulative[i] > 45
               nToRadialRow := aPoints[ 3 ]
               nToRadialCol := aPoints[ 4 ] - Round((aLabelCumulative[i] - 45) / 45 * ( aPoints[ 4 ] - aPoints[ 9 ]),0)
            case aLabelCumulative[i] <= 135 .and. aLabelCumulative[i] > 90
               nToRadialRow := aPoints[ 1 ]
               nToRadialCol := aPoints[ 9 ] - Round((aLabelCumulative[i] - 90) / 45 * ( aPoints[ 9 ] - aPoints[ 2 ]),0)
            case aLabelCumulative[i] <= 180 .and. aLabelCumulative[i] > 135
               nToRadialCol := aPoints[ 2 ]
               nToRadialRow := aPoints[ 1 ] + Round((aLabelCumulative[i] - 135) / 45 * ( aPoints[ 10 ] - aPoints[ 1 ]),0)
            case aLabelCumulative[i] <= 225 .and. aLabelCumulative[i] > 180
               nToRadialCol := aPoints[ 2 ]
               nToRadialRow := aPoints[ 10 ] + Round((aLabelCumulative[i] - 180) / 45 * ( aPoints[ 7 ] - aPoints[ 10 ]),0)
            case aLabelCumulative[i] <= 270 .and. aLabelCumulative[i] > 225
               nToRadialRow := aPoints[ 7 ]
               nToRadialCol := aPoints[ 8 ] + Round((aLabelCumulative[i] - 225) / 45 * ( aPoints[ 12 ] - aPoints[ 8 ]),0)
            case aLabelCumulative[i] <= 315 .and. aLabelCumulative[i] > 270
               nToRadialRow := aPoints[ 7 ]
               nToRadialCol := aPoints[ 12 ] + Round((aLabelCumulative[i] - 270) / 45 * ( aPoints[ 6 ] - aPoints[ 12 ]),0)
            case aLabelCumulative[i] <= 360 .and. aLabelCumulative[i] > 315
               nToRadialCol := aPoints[ 6 ]
               nToRadialRow := aPoints[ 5 ] - Round((aLabelCumulative[i] - 315) / 45 * (aPoints[ 5 ] - aPoints[ 13 ]),0)
            endcase
            BT_DrawText ( ::hDC, nToRadialRow, nToRadialCol, cValue, ::aLegendFont[ 1 ], ::aLegendFont[ 2 ], ::aTextColor, ::aColors[ i ], if( ::aLegendFont[ 3 ], BT_TEXT_BOLD, nil ), BT_TEXT_CENTER)
            nFromRadialRow := nToRadialRow
            nFromRadialCol := nToRadialCol
         next i
         nChartSize := nChartSize - nPieWidth
         nChartSize := nChartSize - nHoleSize
         aPoints := ::FindPieCoordinates( nChartSize )
      next j   
   endif
return nil

Method DrawSunBurstGraph Class GraphPlus
   LOCAL nFromRadialRow, nFromRadialCol, nToRadialRow, nToRadialCol
   LOCAL aDegrees := {}
   LOCAL aCumulative := {}
   LOCAL j,i,sum := 0, lsum := 0
   LOCAL shadowcolor
   LOCAL previous_cumulative
   local aSerieValues := {}
   local nChartSize := 0
   local aLabelCumulative := {}
   local ser_sum := 0
   local nCircles := 0
   local nCircleRatio := 0
   local nPieWidth := 0
   local nHoleSize := 0
   local nCenterRow := 0
   local nCenterCol := 0
   local aPoints := {}
   local nOriginalSize := 0
   local aColor := {}, aHSL := {}
   local aLevelValue := {}
   local lLighter := .f.
   local aLegends := {}, aColors := {}
   ::nLevels := hmg_len( ::aData[ 1 ] ) - 1
   nCircles := ::nLevels
   nCircleRatio := nCircles + ( nCircles * ::nBarGapRatio )
   // convert data
   ::FillSunBurstData()
   ::InitColors()
   if ::lShowLegends   // now draw legends after finding legends
      for j := ::nLevels to 1 step -1
         asize( aLegends, 0 )
         asize( aColors, 0 )
         for i := 1 to hmg_len( ::aLevelHeaders[ j ] )
            aadd( aLegends, ::aLevelHeaders[ j, i ] )
            aLevelValue := ::aLevelValues[ j, i ]
            aColor := ::aColors[ aLevelValue[ 1 ] ]
            aadd( aColors, aColor )
         next i   
         ::DrawSunBurstLegends( aclone( aLegends ), aclone( aColors ) )
      next j   
   endif
   ::nLeft += 25
   ::nRight -= 25
   ::nBottom -= 25
   ::nTop += 25
   nChartSize :=  min( ::nRight - ::nLeft, ::nBottom - ::nTop ) 
   nOriginalSize := nChartSize
   nPieWidth := nChartSize / nCircleRatio
   nHoleSize := nPieWidth * ::nBarGapRatio
   if hmg_len( ::aData ) == 0
      return nil
   endif
   for j := hmg_len( ::aLevelValues ) to 1 step -1
      asize( aSerieValues, 0 )
      asize( aDegrees, 0 )
      asize( aCumulative, 0 )
      asize( aLabelCumulative, 0 )
      for i := 1 to hmg_len( ::aLevelValues[ j ] )
         aLevelValue := ::aLevelValues[ j, i ]
         aadd( aSerieValues, aLevelValue[ 2 ] )
      next i  
      ser_sum := 0
      for i := 1 to HMG_LEN(aSerieValues)
         ser_sum += aSerieValues[i]
      next i
      for i := 1 to HMG_LEN(aSerieValues)
         aadd(aDegrees,Round(aSerieValues[i]/ser_sum * 360,2))
      next i
      sum := 0
      for i := 1 to HMG_LEN(aDegrees)
         sum += aDegrees[i]
      next i
      if sum <> 360
         aDegrees[HMG_LEN(aDegrees)] := aDegrees[HMG_LEN(aDegrees)] + (360 - sum)
      endif

      sum := 0
      lSum := 0
      for i := 1 to HMG_LEN(aDegrees)
         sum := sum + aDegrees[i]
         aadd(aCumulative,sum)
         lSum := lSum + ( aDegrees[i] / 2 )
         aadd( aLabelCumulative, lSum )
         lSum := lSum + ( aDegrees[i] / 2 )
      next i
      for i := 1 to hmg_len( aCumulative )
         if aCumulative[ i ] > 90
            aCumulative[ i ] := 450 - aCumulative[ i ]
         else
            aCumulative[ i ] := 90 - aCumulative[ i ]
         endif
         if aLabelCumulative[ i ] > 90
            aLabelCumulative[ i ] := 450 - aLabelCumulative[ i ]
         else
            aLabelCumulative[ i ] := 90 - aLabelCumulative[ i ]
         endif
      next i   
      previous_cumulative := -1
      aPoints := ::FindPieCoordinates( nChartSize )
      nFromRadialRow := aPoints[ 3 ] // toprightrow
      nFromRadialCol := aPoints[ 9 ] // middletopcol
      for i := 1 to HMG_LEN(aCumulative)
         if aCumulative[i] == previous_cumulative
            loop
         endif
         previous_cumulative := aCumulative[i]
         do case
         case aCumulative[i] <= 45
            nToRadialCol := aPoints[ 14 ] // middlerightcol
            nToRadialRow := aPoints[ 13 ] - Round( aCumulative[i] / 45 * ( aPoints[ 13 ] - aPoints[ 3 ] ),0)
         case aCumulative[i] <= 90 .and. aCumulative[i] > 45
            nToRadialRow := aPoints[ 3 ]
            nToRadialCol := aPoints[ 4 ] - Round((aCumulative[i] - 45) / 45 * ( aPoints[ 4 ] - aPoints[ 9 ]),0)
         case aCumulative[i] <= 135 .and. aCumulative[i] > 90
            nToRadialRow := aPoints[ 1 ]
            nToRadialCol := aPoints[ 9 ] - Round((aCumulative[i] - 90) / 45 * ( aPoints[ 9 ] - aPoints[ 2 ]),0)
         case aCumulative[i] <= 180 .and. aCumulative[i] > 135
            nToRadialCol := aPoints[ 2 ]
            nToRadialRow := aPoints[ 1 ] + Round((aCumulative[i] - 135) / 45 * ( aPoints[ 10 ] - aPoints[ 1 ]),0)
         case aCumulative[i] <= 225 .and. aCumulative[i] > 180
            nToRadialCol := aPoints[ 2 ]
            nToRadialRow := aPoints[ 10 ] + Round((aCumulative[i] - 180) / 45 * ( aPoints[ 7 ] - aPoints[ 10 ]),0)
         case aCumulative[i] <= 270 .and. aCumulative[i] > 225
            nToRadialRow := aPoints[ 7 ]
            nToRadialCol := aPoints[ 8 ] + Round((aCumulative[i] - 225) / 45 * ( aPoints[ 12 ] - aPoints[ 8 ]),0)
         case aCumulative[i] <= 315 .and. aCumulative[i] > 270
            nToRadialRow := aPoints[ 7 ]
            nToRadialCol := aPoints[ 12 ] + Round((aCumulative[i] - 270) / 45 * ( aPoints[ 6 ] - aPoints[ 12 ]),0)
         case aCumulative[i] <= 360 .and. aCumulative[i] > 315
            nToRadialCol := aPoints[ 6 ]
            nToRadialRow := aPoints[ 5 ] - Round((aCumulative[i] - 315) / 45 * (aPoints[ 5 ] - aPoints[ 13 ]),0)
         endcase
         // draw pie
         aLevelValue := ::aLevelValues[ j, i ]
         aColor := ::aColors[ aLevelValue[ 1 ] ]
         BT_DrawPie ( ::hDC, aPoints[ 1 ], aPoints[ 2 ], aPoints[ 5 ], aPoints[ 4 ], nToRadialRow, nToRadialCol, nFromRadialRow, nFromRadialCol, ::aBackColor, ::nPieGap, aColor )
         nFromRadialRow := nToRadialRow
         nFromRadialCol := nToRadialCol
      next i
      nChartSize := nChartSize - nPieWidth
      aPoints := ::FindPieCoordinates( nChartSize )
      // draw the hole!
      nCenterRow := aPoints[ 10 ] - nChartSize / 2 
      nCenterCol := aPoints[ 9 ] - nChartSize / 2 
      BT_DrawFillEllipse ( ::hDC, nCenterRow, nCenterCol, nChartSize, nChartSize, ::aBackColor, ::aBackColor, 1 )
      
      nChartSize := nChartSize - nHoleSize
      aPoints := ::FindPieCoordinates( nChartSize )
   next j   
   if ::lDataLabels
      nChartSize := nOriginalSize
      for j := hmg_len( ::aLevelValues ) to 1 step -1
         asize( aSerieValues, 0 )
         asize( aDegrees, 0 )
         asize( aCumulative, 0 )
         asize( aLabelCumulative, 0 )
         for i := 1 to hmg_len( ::aLevelValues[ j ] )
            aLevelValue := ::aLevelValues[ j, i ]
            aadd( aSerieValues, aLevelValue[ 2 ] )
         next i  
         ser_sum := 0
         for i := 1 to HMG_LEN(aSerieValues)
            ser_sum += aSerieValues[i]
         next i
         for i := 1 to HMG_LEN(aSerieValues)
            aadd(aDegrees,Round(aSerieValues[i]/ser_sum * 360,2))
         next i
         sum := 0
         for i := 1 to HMG_LEN(aDegrees)
            sum += aDegrees[i]
         next i
         if sum <> 360
            aDegrees[HMG_LEN(aDegrees)] := aDegrees[HMG_LEN(aDegrees)] + (360 - sum)
         endif

         sum := 0
         lSum := 0
         for i := 1 to HMG_LEN(aDegrees)
            sum := sum + aDegrees[i]
            aadd(aCumulative,sum)
            lSum := lSum + ( aDegrees[i] / 2 )
            aadd( aLabelCumulative, lSum )
            lSum := lSum + ( aDegrees[i] / 2 )
         next i
         for i := 1 to hmg_len( aCumulative )
            if aCumulative[ i ] > 90
               aCumulative[ i ] := 450 - aCumulative[ i ]
            else
               aCumulative[ i ] := 90 - aCumulative[ i ]
            endif
            if aLabelCumulative[ i ] > 90
               aLabelCumulative[ i ] := 450 - aLabelCumulative[ i ]
            else
               aLabelCumulative[ i ] := 90 - aLabelCumulative[ i ]
            endif
         next i   
         aPoints := ::FindPieCoordinates( nChartSize - nPieWidth / 2 )
         previous_cumulative := -1
         nFromRadialRow := aPoints[ 3 ] // toprightrow
         nFromRadialCol := aPoints[ 9 ] // middletopcol
         for i := 1 to HMG_LEN(aCumulative)
            if aLabelCumulative[i] == previous_cumulative
               loop
            endif
            cValue := AllTrim( Transform( aSerieValues[ i ], ::cPicture ) ) + ' (' + AllTrim( Transform( aSerieValues[ i ] / ser_sum * 100, '999.99' ) ) + '%)'
            previous_cumulative := aLabelCumulative[i]
            do case
            case aLabelCumulative[i] <= 45
               nToRadialCol := aPoints[ 14 ] // middlerightcol
               nToRadialRow := aPoints[ 13 ] - Round( aLabelCumulative[i] / 45 * ( aPoints[ 13 ] - aPoints[ 3 ] ),0)
            case aLabelCumulative[i] <= 90 .and. aLabelCumulative[i] > 45
               nToRadialRow := aPoints[ 3 ]
               nToRadialCol := aPoints[ 4 ] - Round((aLabelCumulative[i] - 45) / 45 * ( aPoints[ 4 ] - aPoints[ 9 ]),0)
            case aLabelCumulative[i] <= 135 .and. aLabelCumulative[i] > 90
               nToRadialRow := aPoints[ 1 ]
               nToRadialCol := aPoints[ 9 ] - Round((aLabelCumulative[i] - 90) / 45 * ( aPoints[ 9 ] - aPoints[ 2 ]),0)
            case aLabelCumulative[i] <= 180 .and. aLabelCumulative[i] > 135
               nToRadialCol := aPoints[ 2 ]
               nToRadialRow := aPoints[ 1 ] + Round((aLabelCumulative[i] - 135) / 45 * ( aPoints[ 10 ] - aPoints[ 1 ]),0)
            case aLabelCumulative[i] <= 225 .and. aLabelCumulative[i] > 180
               nToRadialCol := aPoints[ 2 ]
               nToRadialRow := aPoints[ 10 ] + Round((aLabelCumulative[i] - 180) / 45 * ( aPoints[ 7 ] - aPoints[ 10 ]),0)
            case aLabelCumulative[i] <= 270 .and. aLabelCumulative[i] > 225
               nToRadialRow := aPoints[ 7 ]
               nToRadialCol := aPoints[ 8 ] + Round((aLabelCumulative[i] - 225) / 45 * ( aPoints[ 12 ] - aPoints[ 8 ]),0)
            case aLabelCumulative[i] <= 315 .and. aLabelCumulative[i] > 270
               nToRadialRow := aPoints[ 7 ]
               nToRadialCol := aPoints[ 12 ] + Round((aLabelCumulative[i] - 270) / 45 * ( aPoints[ 6 ] - aPoints[ 12 ]),0)
            case aLabelCumulative[i] <= 360 .and. aLabelCumulative[i] > 315
               nToRadialCol := aPoints[ 6 ]
               nToRadialRow := aPoints[ 5 ] - Round((aLabelCumulative[i] - 315) / 45 * (aPoints[ 5 ] - aPoints[ 13 ]),0)
            endcase
            aLevelValue := ::aLevelValues[ j, i ]
            aColor := ::aColors[ aLevelValue[ 1 ] ]
            BT_DrawText ( ::hDC, nToRadialRow, nToRadialCol, cValue, ::aLegendFont[ 1 ], ::aLegendFont[ 2 ], ::aTextColor, aColor, if( ::aLegendFont[ 3 ], BT_TEXT_BOLD, nil ), BT_TEXT_CENTER)
            nFromRadialRow := nToRadialRow
            nFromRadialCol := nToRadialCol
         next i
         nChartSize := nChartSize - nPieWidth
         nChartSize := nChartSize - nHoleSize
         aPoints := ::FindPieCoordinates( nChartSize )
      next j   
   endif
return nil

Method DrawWaterfallGraph Class GraphPlus
   local nCategories := 0
   local nRow := 0, nCol := 0, nData := 0
   local nTotalBars := 0
   local nAvailableWidth := 0
   local nTotalBarRatio := 0
   local nCategoryRatio := 0
   local nTotalBarSize := 0

   local nBarWidth := 0
   local nCategoryWidth := 0
   local nFactor := 0
   local nBarGap := 0
   local aColor := {}
   
   if hmg_len( ::aData ) > 0
      if valtype( ::aData[ 1 ] ) == 'A'
         nCategories := hmg_len( ::aData[ 1 ] )
      endif
   endif
   if nCategories == 0
      return nil
   endif
   nTotalBars := nCategories 
   nAvailableWidth := ::nRight - ::nLeft
   nCategoryRatio := nCategories * ::BarGapRatio // n Times of Bar size for each category
   nTotalBarRatio := nTotalBars + nCategoryRatio

   nBarWidth := nAvailableWidth / nTotalBarRatio // 100% width
   nCategoryWidth := nBarWidth * ::nBarGapRatio
   
   nCol := ::nLeft + ( nCategoryWidth / 2 )
   nRow := ::nZeroPos
   nMaxHeight := 0
   for i := 1 to hmg_len( ::aData[ 1 ] )
      nData := ::aData[ 1, i ]
      nBarHeight := int( ( ::nBottom - ::nTop ) * ( abs( nData ) / ::nRange ) )
      if ::aData[ 2, i ] // totalitem
         aColor := ::aColors[ 3 ] // total color
         if nData > 0
            // positive
            BT_DrawFillRectangle ( ::hDC, ::nZeroPos - nBarHeight, nCol, nBarWidth, nBarHeight, aColor, aColor, 1 )
            if ::lDataLabels
               BT_DrawText ( ::hDC, ::nZeroPos - nBarHeight + 5, nCol + (nBarWidth / 2 ), alltrim( transform( nData, ::cPicture ) ), ::aLegendFont[ 1 ], ::aLegendFont[ 2 ], ::aTextColor, aColor, if( ::aLegendFont[ 3 ], BT_TEXT_BOLD, nil ), BT_TEXT_CENTER )
            endif
            nRow := ::nZeroPos - nBarHeight // this row becomes the starting row for next increase/decrease
         else
            // negative
            BT_DrawFillRectangle ( ::hDC, ::nZeroPos, nCol, nBarWidth, nBarHeight, aColor, aColor, 1 )
            if ::lDataLabels
               BT_DrawText ( ::hDC, ::nZeroPos + nBarHeight - ::nLabelHeight, nCol + (nBarWidth / 2 ), if( nData < 0, '-', '' ) + alltrim( transform( abs( nData ), ::cPicture ) ), ::aLegendFont[ 1 ], ::aLegendFont[ 2 ], ::aTextColor, aColor, if( ::aLegendFont[ 3 ], BT_TEXT_BOLD, nil ), BT_TEXT_CENTER )
            endif
            nRow := ::nZeroPos + nBarHeight // this row becomes the starting row for next increase/decrease
         endif
      else
         if ::aData[ 1, i ] > 0.0 // increase
            aColor := ::aColors[ 1 ] // color for increase
            BT_DrawFillRectangle ( ::hDC, nRow - nBarHeight, nCol, nBarWidth, nBarHeight, aColor, aColor, 1 )
            if ::lDataLabels
               BT_DrawText ( ::hDC, nRow - nBarHeight + 5, nCol + (nBarWidth / 2 ), alltrim( transform( nData, ::cPicture ) ), ::aLegendFont[ 1 ], ::aLegendFont[ 2 ], ::aTextColor, aColor, if( ::aLegendFont[ 3 ], BT_TEXT_BOLD, nil ), BT_TEXT_CENTER )
            endif
            nRow := nRow - nBarHeight // this row becomes starting point for next increase/decrease
         else // decrease         
            aColor := ::aColors[ 2 ] // color for increase
            BT_DrawFillRectangle ( ::hDC, nRow, nCol, nBarWidth, nBarHeight, aColor, aColor, 1 )
            if ::lDataLabels
               BT_DrawText ( ::hDC, nRow + nBarHeight - ::nLabelHeight, nCol + (nBarWidth / 2 ), '-' + alltrim( transform( abs( nData ), ::cPicture ) ), ::aLegendFont[ 1 ], ::aLegendFont[ 2 ], ::aTextColor, aColor, if( ::aLegendFont[ 3 ], BT_TEXT_BOLD, nil ), BT_TEXT_CENTER )
            endif
            nRow := nRow + nBarHeight // this row becomes starting point for next          
         endif
      endif   
      nCol := nCol + nBarWidth 
      nCol := nCol + nCategoryWidth
   next i
return nil

Method DrawTreeMapGraph Class GraphPlus
   LOCAL nFromRadialRow, nFromRadialCol, nToRadialRow, nToRadialCol
   LOCAL aDegrees := {}
   LOCAL aCumulative := {}
   LOCAL j,i,k,sum := 0, lsum := 0
   LOCAL shadowcolor
   LOCAL previous_cumulative
   local aSerieValues := {}
   local nChartSize := 0
   local aLabelCumulative := {}
   local ser_sum := 0
   local nCircles := 0
   local nCircleRatio := 0
   local nPieWidth := 0
   local nHoleSize := 0
   local nCenterRow := 0
   local nCenterCol := 0
   local aPoints := {}
   local nOriginalSize := 0
   local aColor := {}, aHSL := {}
   local aLevelValue := {}, aLevelHeadValue := {}
   local lLighter := .f.
   local aLegends := {}, aColors := {}
   local nStart := 0, cLabel := ''
   ::nLevels := hmg_len( ::aData[ 1 ] ) - 1
   nCircles := ::nLevels
   nCircleRatio := nCircles + ( nCircles * ::nBarGapRatio )
   // convert data
   ::FillTreeMapData()
   ::InitColors()
   if ::lShowLegends   
      ::DrawLegends()
   endif
   ::nLeft += 25
   ::nRight -= 25
   ::nBottom -= 25
   ::nTop += 25
   nChartSize :=  min( ::nRight - ::nLeft, ::nBottom - ::nTop ) 
   nOriginalSize := nChartSize
   nPieWidth := nChartSize / nCircleRatio
   nHoleSize := nPieWidth * ::nBarGapRatio
   if hmg_len( ::aData ) == 0
      return nil
   endif
   aPositions := { ::nTop, ::nLeft, ::nRight - ::nLeft, ::nBottom - ::nTop }
   for j := 1 to hmg_len( ::aLevelValues )
      asize( ::aTreeMapData, 0 )
      asize( aDegrees, 0 )
      asize( aCumulative, 0 )
      asize( aLabelCumulative, 0 )
      asize( ::aTreeMapRect, 0 )
      if j > 1  // sub level
         nHead := 1
         aLevelHeadValue := ::aLevelValues[ j - 1, nHead ]
         cCurHeader := aLevelHeadValue[ 3 ]
         nStart := 1
         nCount := 0
         aPositions := aLevelHeadValue[ 5 ]
         aColor := ::aColors[ ::aLevelValues[ j, 1 ][ 1 ] ]
         for i := 1 to hmg_len( ::aLevelValues[ j ] )
            aLevelValue := ::aLevelValues[ j, i ]
            if aLevelValue[ 4 ] == cCurHeader
               aadd( ::aTreeMapData, aLevelValue[ 2 ] )
               aadd( ::aTreeMapRect, { 0, 0, 0, 0 } )
               nCount++
            else // same level another header
               // drawtreemap now
               ::StripTreemap( aPositions[ 1 ], aPositions[ 2 ], aPositions[ 3 ], aPositions[ 4 ] )
               for k := nStart to nStart + nCount - 1
                  aadd( ::aLevelValues[ j, k ], aclone( ::aTreeMapRect[ k - nStart + 1 ] ) )
               next k   
               for k := 1 to hmg_len( ::aTreeMapRect )
                  nRow := ::aTreeMapRect[ k, 1 ]
                  nCol := ::aTreeMapRect[ k, 2 ]
                  nWidth := ::aTreeMapRect[ k, 3 ]
                  nHeight := ::aTreeMapRect[ k, 4 ]
                  BT_DrawFillRectangle ( ::hDC, nRow, nCol, nWidth, nHeight, aColor, ::aBackColor, 2 )
                  // show data label if required
                  if j == hmg_len( ::aLevelValues ) // only in last level
                     if ::lDataLabels // show labels
                        cLabel := ::aLevelValues[ j, nStart + k - 1 ][ 3 ]
                        nFound := 0
                        nFound := HMG_AT( ': ', cLabel )
                        if nFound > 0
                           cLabel := hmg_substr( cLabel, nFound + 2 )
                        endif   
                        BT_DrawTextEx (::hDC, nRow + 2, nCol + 2, nWidth - 4, nHeight, alltrim( cLabel ), ::aLegendFont[ 1 ], ::aLegendFont[ 2 ], ::aTextColor, aColor, if( ::aLegendFont[ 3 ], BT_TEXT_BOLD, nil ))
                     endif
                  endif   
               next k   
               nStart := nStart + nCount
               nCount := 0
               nHead++
               asize( ::aTreeMapData, 0 )
               asize( aDegrees, 0 )
               asize( aCumulative, 0 )
               asize( aLabelCumulative, 0 )
               asize( ::aTreeMapRect, 0 )
               if nHead <= hmg_len( ::aLevelValues[ j - 1 ] )
                  aLevelHeadValue := ::aLevelValues[ j - 1, nHead ]
                  cCurHeader := aLevelHeadValue[ 3 ]
                  aPositions := aLevelHeadValue[ 5 ]
               endif   
               aadd( ::aTreeMapData, aLevelValue[ 2 ] )
               aadd( ::aTreeMapRect, { 0, 0, 0, 0 } )
               nCount++
               aColor := ::aColors[ ::aLevelValues[ j, i ][ 1 ] ]
            endif   
         next i
         // for the last one
         ::StripTreemap( aPositions[ 1 ], aPositions[ 2 ], aPositions[ 3 ], aPositions[ 4 ] )
         for k := nStart to nStart + nCount - 1
            aadd( ::aLevelValues[ j, k ], aclone( ::aTreeMapRect[ k - nStart + 1 ] ) )
         next k   
         for k := 1 to hmg_len( ::aTreeMapRect )
            nRow := ::aTreeMapRect[ k, 1 ]
            nCol := ::aTreeMapRect[ k, 2 ]
            nWidth := ::aTreeMapRect[ k, 3 ]
            nHeight := ::aTreeMapRect[ k, 4 ]
            BT_DrawFillRectangle ( ::hDC, nRow, nCol, nWidth, nHeight, aColor, ::aBackColor, 2 )
            if j == hmg_len( ::aLevelValues ) // only in last level
               if ::lDataLabels // show labels
                  cLabel := ::aLevelValues[ j, nStart + k - 1 ][ 3 ]
                  nFound := 0
                  nFound := HMG_AT( ': ', cLabel )
                  if nFound > 0
                     cLabel := hmg_substr( cLabel, nFound + 2 )
                  endif   
                  BT_DrawTextEx (::hDC, nRow + 2, nCol + 2, nWidth - 4, nHeight, alltrim( cLabel ), ::aLegendFont[ 1 ], ::aLegendFont[ 2 ], ::aTextColor, aColor, if( ::aLegendFont[ 3 ], BT_TEXT_BOLD, nil ))
               endif
            endif   
         next k   
      else         
         for i := 1 to hmg_len( ::aLevelValues[ j ] )
            aLevelValue := ::aLevelValues[ j, i ]
            aadd( ::aTreeMapData, aLevelValue[ 2 ] )
            aadd( ::aTreeMapRect, { 0, 0, 0, 0 } )
         next i  
         ::StripTreemap( aPositions[ 1 ], aPositions[ 2 ], aPositions[ 3 ], aPositions[ 4 ] )
         for i := 1 to hmg_len( ::aLevelValues[ j ] )
            aadd( ::aLevelValues[ j, i ], aclone( ::aTreeMapRect[ i ] ) )
         next i
         for k := 1 to hmg_len( ::aTreeMapRect )
            nRow := ::aTreeMapRect[ k, 1 ]
            nCol := ::aTreeMapRect[ k, 2 ]
            nWidth := ::aTreeMapRect[ k, 3 ]
            nHeight := ::aTreeMapRect[ k, 4 ]
            BT_DrawFillRectangle ( ::hDC, nRow, nCol, nWidth, nHeight, ::aColors[ ::aLevelValues[ j, k ][ 1 ] ], ::aBackColor, 2 )
            if j == hmg_len( ::aLevelValues ) // if only one level
               if ::lDataLabels // show labels
                  cLabel := ::aLevelValues[ j, k ][ 3 ]
                  nFound := 0
                  nFound := HMG_AT( ': ', cLabel )
                  if nFound > 0
                     cLabel := hmg_substr( cLabel, nFound + 2 )
                  endif   
                  BT_DrawTextEx (::hDC, nRow + 2, nCol + 2, nWidth - 4, nHeight, alltrim( cLabel ), ::aLegendFont[ 1 ], ::aLegendFont[ 2 ], ::aTextColor, ::aColors[ ::aLevelValues[ j, k ][ 1 ] ], if( ::aLegendFont[ 3 ], BT_TEXT_BOLD, nil ))
               endif
            endif   
         next k   
      endif   
   next j   
   return nil
   
   for j := hmg_len( ::aLevelValues ) to 1 step -1
      asize( aSerieValues, 0 )
      asize( aDegrees, 0 )
      asize( aCumulative, 0 )
      asize( aLabelCumulative, 0 )
      for i := 1 to hmg_len( ::aLevelValues[ j ] )
         aLevelValue := ::aLevelValues[ j, i ]
         aadd( aSerieValues, aLevelValue[ 2 ] )
      next i  
      ser_sum := 0
      for i := 1 to HMG_LEN(aSerieValues)
         ser_sum += aSerieValues[i]
      next i
      for i := 1 to HMG_LEN(aSerieValues)
         aadd(aDegrees,Round(aSerieValues[i]/ser_sum * 360,2))
      next i
      sum := 0
      for i := 1 to HMG_LEN(aDegrees)
         sum += aDegrees[i]
      next i
      if sum <> 360
         aDegrees[HMG_LEN(aDegrees)] := aDegrees[HMG_LEN(aDegrees)] + (360 - sum)
      endif

      sum := 0
      lSum := 0
      for i := 1 to HMG_LEN(aDegrees)
         sum := sum + aDegrees[i]
         aadd(aCumulative,sum)
         lSum := lSum + ( aDegrees[i] / 2 )
         aadd( aLabelCumulative, lSum )
         lSum := lSum + ( aDegrees[i] / 2 )
      next i
      for i := 1 to hmg_len( aCumulative )
         if aCumulative[ i ] > 90
            aCumulative[ i ] := 450 - aCumulative[ i ]
         else
            aCumulative[ i ] := 90 - aCumulative[ i ]
         endif
         if aLabelCumulative[ i ] > 90
            aLabelCumulative[ i ] := 450 - aLabelCumulative[ i ]
         else
            aLabelCumulative[ i ] := 90 - aLabelCumulative[ i ]
         endif
      next i   
      previous_cumulative := -1
      aPoints := ::FindPieCoordinates( nChartSize )
      nFromRadialRow := aPoints[ 3 ] // toprightrow
      nFromRadialCol := aPoints[ 9 ] // middletopcol
      for i := 1 to HMG_LEN(aCumulative)
         if aCumulative[i] == previous_cumulative
            loop
         endif
         previous_cumulative := aCumulative[i]
         do case
         case aCumulative[i] <= 45
            nToRadialCol := aPoints[ 14 ] // middlerightcol
            nToRadialRow := aPoints[ 13 ] - Round( aCumulative[i] / 45 * ( aPoints[ 13 ] - aPoints[ 3 ] ),0)
         case aCumulative[i] <= 90 .and. aCumulative[i] > 45
            nToRadialRow := aPoints[ 3 ]
            nToRadialCol := aPoints[ 4 ] - Round((aCumulative[i] - 45) / 45 * ( aPoints[ 4 ] - aPoints[ 9 ]),0)
         case aCumulative[i] <= 135 .and. aCumulative[i] > 90
            nToRadialRow := aPoints[ 1 ]
            nToRadialCol := aPoints[ 9 ] - Round((aCumulative[i] - 90) / 45 * ( aPoints[ 9 ] - aPoints[ 2 ]),0)
         case aCumulative[i] <= 180 .and. aCumulative[i] > 135
            nToRadialCol := aPoints[ 2 ]
            nToRadialRow := aPoints[ 1 ] + Round((aCumulative[i] - 135) / 45 * ( aPoints[ 10 ] - aPoints[ 1 ]),0)
         case aCumulative[i] <= 225 .and. aCumulative[i] > 180
            nToRadialCol := aPoints[ 2 ]
            nToRadialRow := aPoints[ 10 ] + Round((aCumulative[i] - 180) / 45 * ( aPoints[ 7 ] - aPoints[ 10 ]),0)
         case aCumulative[i] <= 270 .and. aCumulative[i] > 225
            nToRadialRow := aPoints[ 7 ]
            nToRadialCol := aPoints[ 8 ] + Round((aCumulative[i] - 225) / 45 * ( aPoints[ 12 ] - aPoints[ 8 ]),0)
         case aCumulative[i] <= 315 .and. aCumulative[i] > 270
            nToRadialRow := aPoints[ 7 ]
            nToRadialCol := aPoints[ 12 ] + Round((aCumulative[i] - 270) / 45 * ( aPoints[ 6 ] - aPoints[ 12 ]),0)
         case aCumulative[i] <= 360 .and. aCumulative[i] > 315
            nToRadialCol := aPoints[ 6 ]
            nToRadialRow := aPoints[ 5 ] - Round((aCumulative[i] - 315) / 45 * (aPoints[ 5 ] - aPoints[ 13 ]),0)
         endcase
         // draw pie
         aLevelValue := ::aLevelValues[ j, i ]
         aColor := ::aColors[ aLevelValue[ 1 ] ]
         BT_DrawPie ( ::hDC, aPoints[ 1 ], aPoints[ 2 ], aPoints[ 5 ], aPoints[ 4 ], nToRadialRow, nToRadialCol, nFromRadialRow, nFromRadialCol, ::aBackColor, ::nPieGap, aColor )
         nFromRadialRow := nToRadialRow
         nFromRadialCol := nToRadialCol
      next i
      nChartSize := nChartSize - nPieWidth
      aPoints := ::FindPieCoordinates( nChartSize )
      // draw the hole!
      nCenterRow := aPoints[ 10 ] - nChartSize / 2 
      nCenterCol := aPoints[ 9 ] - nChartSize / 2 
      BT_DrawFillEllipse ( ::hDC, nCenterRow, nCenterCol, nChartSize, nChartSize, ::aBackColor, ::aBackColor, 1 )
      
      nChartSize := nChartSize - nHoleSize
      aPoints := ::FindPieCoordinates( nChartSize )
   next j   
   if ::lDataLabels
      nChartSize := nOriginalSize
      for j := hmg_len( ::aLevelValues ) to 1 step -1
         asize( aSerieValues, 0 )
         asize( aDegrees, 0 )
         asize( aCumulative, 0 )
         asize( aLabelCumulative, 0 )
         for i := 1 to hmg_len( ::aLevelValues[ j ] )
            aLevelValue := ::aLevelValues[ j, i ]
            aadd( aSerieValues, aLevelValue[ 2 ] )
         next i  
         ser_sum := 0
         for i := 1 to HMG_LEN(aSerieValues)
            ser_sum += aSerieValues[i]
         next i
         for i := 1 to HMG_LEN(aSerieValues)
            aadd(aDegrees,Round(aSerieValues[i]/ser_sum * 360,2))
         next i
         sum := 0
         for i := 1 to HMG_LEN(aDegrees)
            sum += aDegrees[i]
         next i
         if sum <> 360
            aDegrees[HMG_LEN(aDegrees)] := aDegrees[HMG_LEN(aDegrees)] + (360 - sum)
         endif

         sum := 0
         lSum := 0
         for i := 1 to HMG_LEN(aDegrees)
            sum := sum + aDegrees[i]
            aadd(aCumulative,sum)
            lSum := lSum + ( aDegrees[i] / 2 )
            aadd( aLabelCumulative, lSum )
            lSum := lSum + ( aDegrees[i] / 2 )
         next i
         for i := 1 to hmg_len( aCumulative )
            if aCumulative[ i ] > 90
               aCumulative[ i ] := 450 - aCumulative[ i ]
            else
               aCumulative[ i ] := 90 - aCumulative[ i ]
            endif
            if aLabelCumulative[ i ] > 90
               aLabelCumulative[ i ] := 450 - aLabelCumulative[ i ]
            else
               aLabelCumulative[ i ] := 90 - aLabelCumulative[ i ]
            endif
         next i   
         aPoints := ::FindPieCoordinates( nChartSize - nPieWidth / 2 )
         previous_cumulative := -1
         nFromRadialRow := aPoints[ 3 ] // toprightrow
         nFromRadialCol := aPoints[ 9 ] // middletopcol
         for i := 1 to HMG_LEN(aCumulative)
            if aLabelCumulative[i] == previous_cumulative
               loop
            endif
            cValue := AllTrim( Transform( aSerieValues[ i ], ::cPicture ) ) + ' (' + AllTrim( Transform( aSerieValues[ i ] / ser_sum * 100, '999.99' ) ) + '%)'
            previous_cumulative := aLabelCumulative[i]
            do case
            case aLabelCumulative[i] <= 45
               nToRadialCol := aPoints[ 14 ] // middlerightcol
               nToRadialRow := aPoints[ 13 ] - Round( aLabelCumulative[i] / 45 * ( aPoints[ 13 ] - aPoints[ 3 ] ),0)
            case aLabelCumulative[i] <= 90 .and. aLabelCumulative[i] > 45
               nToRadialRow := aPoints[ 3 ]
               nToRadialCol := aPoints[ 4 ] - Round((aLabelCumulative[i] - 45) / 45 * ( aPoints[ 4 ] - aPoints[ 9 ]),0)
            case aLabelCumulative[i] <= 135 .and. aLabelCumulative[i] > 90
               nToRadialRow := aPoints[ 1 ]
               nToRadialCol := aPoints[ 9 ] - Round((aLabelCumulative[i] - 90) / 45 * ( aPoints[ 9 ] - aPoints[ 2 ]),0)
            case aLabelCumulative[i] <= 180 .and. aLabelCumulative[i] > 135
               nToRadialCol := aPoints[ 2 ]
               nToRadialRow := aPoints[ 1 ] + Round((aLabelCumulative[i] - 135) / 45 * ( aPoints[ 10 ] - aPoints[ 1 ]),0)
            case aLabelCumulative[i] <= 225 .and. aLabelCumulative[i] > 180
               nToRadialCol := aPoints[ 2 ]
               nToRadialRow := aPoints[ 10 ] + Round((aLabelCumulative[i] - 180) / 45 * ( aPoints[ 7 ] - aPoints[ 10 ]),0)
            case aLabelCumulative[i] <= 270 .and. aLabelCumulative[i] > 225
               nToRadialRow := aPoints[ 7 ]
               nToRadialCol := aPoints[ 8 ] + Round((aLabelCumulative[i] - 225) / 45 * ( aPoints[ 12 ] - aPoints[ 8 ]),0)
            case aLabelCumulative[i] <= 315 .and. aLabelCumulative[i] > 270
               nToRadialRow := aPoints[ 7 ]
               nToRadialCol := aPoints[ 12 ] + Round((aLabelCumulative[i] - 270) / 45 * ( aPoints[ 6 ] - aPoints[ 12 ]),0)
            case aLabelCumulative[i] <= 360 .and. aLabelCumulative[i] > 315
               nToRadialCol := aPoints[ 6 ]
               nToRadialRow := aPoints[ 5 ] - Round((aLabelCumulative[i] - 315) / 45 * (aPoints[ 5 ] - aPoints[ 13 ]),0)
            endcase
            aLevelValue := ::aLevelValues[ j, i ]
            aColor := ::aColors[ aLevelValue[ 1 ] ]
            BT_DrawText ( ::hDC, nToRadialRow, nToRadialCol, cValue, ::aLegendFont[ 1 ], ::aLegendFont[ 2 ], ::aTextColor, aColor, if( ::aLegendFont[ 3 ], BT_TEXT_BOLD, nil ), BT_TEXT_CENTER)
            nFromRadialRow := nToRadialRow
            nFromRadialCol := nToRadialCol
         next i
         nChartSize := nChartSize - nPieWidth
         nChartSize := nChartSize - nHoleSize
         aPoints := ::FindPieCoordinates( nChartSize )
      next j   
   endif
return nil




Method DrawBarGraph Class GraphPlus
   local nCategories := 0
   local nRow := 0, nCol := 0, nData := 0
   local nTotalBars := 0
   local nAvailableHeight := 0
   local nTotalBarRatio := 0
   local nCategoryRatio := 0
   local nTotalBarSize := 0

   local nBarWidth := 0
   local nBarHeight := 0
   local nCategoryHeight := 0
   local nFactor := 0
   local nBarGap := 0
   
   
   if hmg_len( ::aData ) > 0
      if valtype( ::aData[ 1 ] ) == 'A'
         nCategories := hmg_len( ::aData[ 1 ] )
      endif
   endif
   if nCategories == 0
      return nil
   endif
   nTotalBars := nCategories * hmg_len( ::aData )
   nAvailableHeight := ::nBottom - ::nTop
   nTotalBarRatio := nTotalBars * 1 // 100%
   nTotalBarRatio := nTotalBarRatio + ( nTotalBarRatio - nCategories ) * ::nBarGapRatio 
   nCategoryRatio := nCategories * ::nGapWidthRatio // n Times of Bar size for each category
   nTotalBarRatio := nTotalBarRatio + nCategoryRatio

   nBarHeight := nAvailableHeight / nTotalBarRatio // 100% width
   nCategoryHeight := nBarHeight * ::nGapWidthRatio
   nBarGap      := nBarHeight * ::nBarGapRatio

   nRow := ::nTop + ( nCategoryHeight / 2 )
   nCol := ::nZeroPos
   for i := 1 to hmg_len( ::aData[ 1 ] )
      for j := 1 to hmg_len( ::aData )
         nData := ::aData[ j, i ]
         nBarWidth := int( ( ::nRight - ::nLeft ) * ( abs( nData ) / ::nRange ) )
         if nData > 0
            // positive
            BT_DrawFillRectangle ( ::hDC, nRow, nCol, nBarWidth, nBarHeight, ::aColors[ j ], ::aColors[ j ], 1 )
            if ::lDataLabels
               BT_DrawText ( ::hDC, nRow + ( nBarHeight / 2 ), nCol + nBarWidth - 5, alltrim( transform( nData, ::cPicture ) ), ::aLegendFont[ 1 ], ::aLegendFont[ 2 ], ::aTextColor, ::aColors[j], if( ::aLegendFont[ 3 ], BT_TEXT_BOLD, nil ), BT_TEXT_BASELINE + BT_TEXT_RIGHT )
            endif
         else
            // negative
            BT_DrawFillRectangle ( ::hDC, nRow, nCol - nBarWidth, nBarWidth, nBarHeight, ::aColors[ j ], ::aColors[ j ], 1 )
            if ::lDataLabels
               BT_DrawText ( ::hDC, nRow + ( nBarHeight / 2 ), nCol - nBarWidth + 5, if( nData < 0, '-', '' ) + alltrim( transform( abs( nData ), ::cPicture ) ), ::aLegendFont[ 1 ], ::aLegendFont[ 2 ], ::aTextColor, ::aColors[j], if( ::aLegendFont[ 3 ], BT_TEXT_BOLD, nil ), BT_TEXT_BASELINE + BT_TEXT_LEFT )
            endif
         endif
         nRow := nRow + nBarHeight
         if j < hmg_len( ::aData )
            nRow := nRow + nBarGap
         endif   
      next j
      nRow := nRow + nCategoryHeight
   next i
return nil

Method DrawFunnelGraph Class GraphPlus
   local nCategories := 0
   local nRow := 0, nCol := 0, nData := 0
   local nTotalBars := 0
   local nAvailableHeight := 0
   local nTotalBarRatio := 0
   local nCategoryRatio := 0
   local nTotalBarSize := 0

   local nBarWidth := 0
   local nBarHeight := 0
   local nCategoryHeight := 0
   local nFactor := 0
   local nBarGap := 0
   local aFunnelData := aclone( ::aData[ 1 ] )
   local nMax := 0
   nCategories := hmg_len( aFunnelData )
   // find max value
   nMax := 0
   FOR i := 1 TO HMG_LEN( aFunnelData )
      nMax := Max( aFunnelData[ i ], nMax )
   NEXT i
   if nMax <= 0 
      msgstop( 'Funnel can not be drawn for negative or zero values!' )      
      return nil
   endif
   // title height
   if hmg_len( ::cTitle ) > 0 
      aSize := BT_DrawTextSize (::hDC, ::cTitle, ::aTitleFont[ 1 ], ::aTitleFont[ 2 ], if( ::aTitleFont[ 3 ], BT_TEXT_BOLD, nil ))
      ::nTop := ::nTop + aSize[ 2 ]
   endif
   if ::lShowCategories 
      ::DrawCategories()
   endif
   if hmg_len( ::cPicture ) == 0
      ::cPicture := replicate( '9', hmg_len( alltrim( str( nMax ) ) ))
   endif   
   
   ::nRight := ::nRight - 10
   ::nLeft := ::nLeft + 5
   ::nTop := ::nTop + 10
   nTotalBars := hmg_len( aFunnelData )
   nAvailableHeight := ::nBottom - ::nTop
   nTotalBarRatio := nTotalBars * 1 // 100%
   nCategoryRatio := nCategories * ::nGapWidthRatio // n Times of Bar size for each category
   nTotalBarRatio := nTotalBarRatio + nCategoryRatio

   nBarHeight := nAvailableHeight / nTotalBarRatio // 100% width
   nCategoryHeight := nBarHeight * ::nGapWidthRatio
   // draw left vertical line
   BT_DrawLine (::hDC, ::nTop, ::nLeft, ::nBottom, ::nLeft, ::aGridLineColor, ::nLineWidth)
   nWidth := ( ::nRight - ::nLeft ) - 10
   nRow := ::nTop
   for i := 1 to hmg_len( aFunnelData )
      nData := aFunnelData[ i ]
      nBarWidth := nWidth * ( nData / nMax )
      nFromCol := ::nLeft + 5 + ( ( nWidth - nBarWidth ) / 2 )
      if nData > 0 // ( only for non negative )
         BT_DrawFillRectangle ( ::hDC, nRow, nFromCol, nBarWidth, nBarHeight, ::aColors[ 1 ], ::aColors[ 1 ], 1 )
         BT_DrawText ( ::hDC, ( nRow + nBarHeight / 2 ) , ::nLeft + 5 + ( nWidth / 2 ) , alltrim( transform( aFunnelData[ i ], ::cPicture ) ), ::aLegendFont[ 1 ], ::aLegendFont[ 2 ], ::aBackColor, ::aColors[ 1 ], if( ::aLegendFont[ 3 ], BT_TEXT_BOLD, nil ), BT_TEXT_BASELINE + BT_TEXT_CENTER)
      endif
      nRow := nRow + nBarHeight + nCategoryHeight
   next i
   // title
   if hmg_len( ::cTitle ) > 0
      BT_DrawText ( ::hDC, 5, ::nLeft + 5 + ( nWidth / 2 ), ::cTitle, ::aTitleFont[ 1 ], ::aTitleFont[ 2 ], ::aTitleColor, ::aBackColor, if( ::aTitleFont[ 3 ], BT_TEXT_BOLD, nil ), BT_TEXT_CENTER )
   endif
return nil

Method DrawStackedColumnGraph Class GraphPlus
   local nCategories := 0
   local nRow := 0, nCol := 0, nData := 0
   local nTotalBars := 0
   local nAvailableWidth := 0
   local nTotalBarRatio := 0
   local nCategoryRatio := 0
   local nTotalBarSize := 0

   local nBarWidth := 0
   local nCategoryWidth := 0
   local nFactor := 0
   local nBarGap := 0
   local nNegativeRow := 0, nPositiveRow := 0
   
   
   if hmg_len( ::aData ) > 0
      if valtype( ::aData[ 1 ] ) == 'A'
         nCategories := hmg_len( ::aData[ 1 ] )
      endif
   endif
   if nCategories == 0
      return nil
   endif
   nTotalBars := nCategories 
   nAvailableWidth := ::nRight - ::nLeft
   nTotalBarRatio := nTotalBars * 1 // 100%
   nCategoryRatio := nCategories * ::nGapWidthRatio // n Times of Bar size for each category
   nTotalBarRatio := nTotalBarRatio + nCategoryRatio

   nBarWidth := nAvailableWidth / nTotalBarRatio // 100% width
   nCategoryWidth := nBarWidth * ::nGapWidthRatio
   nBarGap      := nBarWidth * ::nBarGapRatio
   
   nCol := ::nLeft + ( nCategoryWidth / 2 )
   nRow := ::nZeroPos
   nPositiveRow := nRow
   nNegativeRow := nRow
   for i := 1 to hmg_len( ::aData[ 1 ] )
      nPositiveRow := nRow
      nNegativeRow := nRow
      for j := 1 to hmg_len( ::aData )
         nData := ::aData[ j, i ]
         nBarHeight := int( ( ::nBottom - ::nTop ) * ( abs( nData ) / ::nRange ) )
         if nData > 0
            // positive
            BT_DrawFillRectangle ( ::hDC, nPositiveRow - nBarHeight, nCol, nBarWidth, nBarHeight, ::aColors[ j ], ::aColors[ j ], 1 )
            if ::lDataLabels
               BT_DrawText ( ::hDC, nPositiveRow - nBarHeight + 5, nCol + (nBarWidth / 2 ), alltrim( transform( nData, ::cPicture ) ), ::aLegendFont[ 1 ], ::aLegendFont[ 2 ], ::aTextColor, ::aColors[ j ], if( ::aLegendFont[ 3 ], BT_TEXT_BOLD, nil ), BT_TEXT_CENTER )
            endif
            nPositiveRow := nPositiveRow - nBarHeight
         else
            // negative
            BT_DrawFillRectangle ( ::hDC, nNegativeRow, nCol, nBarWidth, nBarHeight, ::aColors[ j ], ::aColors[ j ], 1 )
            if ::lDataLabels
               BT_DrawText ( ::hDC, nNegativeRow + nBarHeight - 5, nCol + (nBarWidth / 2 ), if( nData < 0, '-', '' ) + alltrim( transform( abs( nData ), ::cPicture ) ), ::aLegendFont[ 1 ], ::aLegendFont[ 2 ], ::aTextColor, ::aColors[ j ], if( ::aLegendFont[ 3 ], BT_TEXT_BOLD, nil ), BT_TEXT_BOTTOM + BT_TEXT_CENTER )
            endif
            nNegativeRow := nNegativeRow + nBarHeight
         endif
      next j
      nCol := nCol + nBarWidth + nCategoryWidth
   next i
return nil

Method DrawStackedBarGraph Class GraphPlus
   local nCategories := 0
   local nRow := 0, nCol := 0, nData := 0
   local nTotalBars := 0
   local nAvailableHeight := 0
   local nTotalBarRatio := 0
   local nCategoryRatio := 0
   local nTotalBarSize := 0

   local nBarWidth := 0
   local nBarHeight := 0
   local nCategoryHeight := 0
   local nFactor := 0
   local nBarGap := 0
   local nNegativeCol := 0, nPositiveCol := 0
   
   
   if hmg_len( ::aData ) > 0
      if valtype( ::aData[ 1 ] ) == 'A'
         nCategories := hmg_len( ::aData[ 1 ] )
      endif
   endif
   if nCategories == 0
      return nil
   endif
   nTotalBars := nCategories 
   nAvailableHeight := ::nBottom - ::nTop
   nTotalBarRatio := nTotalBars * 1 // 100%
   nCategoryRatio := nCategories * ::nGapWidthRatio // n Times of Bar size for each category
   nTotalBarRatio := nTotalBarRatio + nCategoryRatio

   nBarHeight := nAvailableHeight / nTotalBarRatio // 100% width
   nCategoryHeight := nBarHeight * ::nGapWidthRatio

   nRow := ::nTop + ( nCategoryHeight / 2 )
   nCol := ::nZeroPos
   for i := 1 to hmg_len( ::aData[ 1 ] )
      nPositiveCol := nCol
      nNegativeCol := nCol
      for j := 1 to hmg_len( ::aData )
         nData := ::aData[ j, i ]
         nBarWidth := int( ( ::nRight - ::nLeft ) * ( abs( nData ) / ::nRange ) )
         if nData > 0
            // positive
            BT_DrawFillRectangle ( ::hDC, nRow, nPositiveCol, nBarWidth, nBarHeight, ::aColors[ j ], ::aColors[ j ], 1 )
            if ::lDataLabels
               BT_DrawText ( ::hDC, nRow + ( nBarHeight / 2 ), nPositiveCol + nBarWidth - 5, alltrim( transform( nData, ::cPicture ) ), ::aLegendFont[ 1 ], ::aLegendFont[ 2 ], ::aTextColor, ::aColors[j], if( ::aLegendFont[ 3 ], BT_TEXT_BOLD, nil ), BT_TEXT_BASELINE + BT_TEXT_RIGHT )
            endif
            nPositiveCol := nPositiveCol + nBarWidth
         else
            // negative
            BT_DrawFillRectangle ( ::hDC, nRow, nNegativeCol - nBarWidth, nBarWidth, nBarHeight, ::aColors[ j ], ::aColors[ j ], 1 )
            if ::lDataLabels
               BT_DrawText ( ::hDC, nRow + ( nBarHeight / 2 ), nNegativeCol - nBarWidth + 5, if( nData < 0, '-', '' ) + alltrim( transform( abs( nData ), ::cPicture ) ), ::aLegendFont[ 1 ], ::aLegendFont[ 2 ], ::aTextColor, ::aColors[j], if( ::aLegendFont[ 3 ], BT_TEXT_BOLD, nil ), BT_TEXT_BASELINE + BT_TEXT_LEFT )
            endif
            nNegativeCol := nNegativeCol - nBarWidth
         endif
      next j
      nRow := nRow + nBarHeight + nCategoryHeight
   next i
return nil

Method DrawAreaGraph Class GraphPlus
   local nCategories := 0
   local nRow := 0, nCol := 0, nData := 0, nFirstCol := 0, nPrevRow := 0, nPrevCol := 0
   local nTotalBars := 0
   local nAvailableWidth := 0
   local nTotalBarRatio := 0
   local nCategoryRatio := 0
   local nTotalBarSize := 0

   local nBarWidth := 0
   local nCategoryWidth := 0
   local nFactor := 0
   local nBarGap := 0
   local aRows := {}
   local aCols := {}
   
   
   if hmg_len( ::aData ) > 0
      if valtype( ::aData[ 1 ] ) == 'A'
         nCategories := hmg_len( ::aData[ 1 ] )
      endif
   endif
   if nCategories == 0
      return nil
   endif
   nAvailableWidth := ::nRight - ::nLeft
   nTotalBarRatio := nCategories - 1 // equal space between categories

   nBarWidth := nAvailableWidth / nTotalBarRatio // 100% width
   nCategoryWidth := nBarWidth 
   nFirstCol := ::nLeft 
   nCol := ::nLeft 
   nRow := ::nZeroPos
   for j := 1 to hmg_len( ::aData )
      nCol := nFirstCol
      asize( aRows, 0 )
      asize( aCols, 0 )
      aadd( aRows, ::nZeroPos )  // starting point at zero
      aadd( aCols, ::nLeft )     // starting point at left
      for i := 1 to hmg_len( ::aData[ 1 ] )
         nData := ::aData[ j, i ]
         nBarHeight := int( ( ::nBottom - ::nTop ) * ( abs( nData ) / ::nRange ) )
         if nData > 0
            if ::lDataLabels
               BT_DrawText ( ::hDC, nRow - nBarHeight - 15 - ::nPointSize / 2 , nCol, alltrim( transform( nData, ::cPicture ) ), ::aLegendFont[ 1 ], ::aLegendFont[ 2 ], ::aTextColor, ::aBackColor, if( ::aLegendFont[ 3 ], BT_TEXT_BOLD, nil ), BT_TEXT_CENTER )
            endif
            nPrevRow := nRow - nBarHeight
            nPrevCol := nCol
         else
            if ::lDataLabels
               BT_DrawText ( ::hDC, nRow + nBarHeight + 3 + ::nPointSize / 2 , nCol, if( nData < 0, '-', '' ) + alltrim( transform( abs( nData ), ::cPicture ) ), ::aLegendFont[ 1 ], ::aLegendFont[ 2 ], ::aTextColor, ::aBackColor, if( ::aLegendFont[ 3 ], BT_TEXT_BOLD, nil ), BT_TEXT_CENTER )
            endif
            nPrevRow := nRow + nBarHeight
            nPrevCol := nCol
         endif
         aadd( aCols, nCol )
         aadd( aRows, nPrevRow )
         nCol := nCol + nCategoryWidth
      next i
      aadd( aCols, ::nRight ) // end point in the right
      aadd( aRows, ::nZeroPos ) // end point zero pos
      BT_DrawPolygon (::hDC, aRows, aCols, ::aColors[ j ], 1, ::aColors[ j ] )
   next j
   
   // data labels...
   if ::lDataLabels
      nCol := ::nLeft 
      nRow := ::nZeroPos
      for j := 1 to hmg_len( ::aData )
         nCol := nFirstCol
         for i := 1 to hmg_len( ::aData[ 1 ] )
            nData := ::aData[ j, i ]
            nBarHeight := int( ( ::nBottom - ::nTop ) * ( abs( nData ) / ::nRange ) )
            if nData > 0 // positive
               BT_DrawText ( ::hDC, nRow - nBarHeight - 15 - ::nPointSize / 2 , nCol, alltrim( transform( nData, ::cPicture ) ), ::aLegendFont[ 1 ], ::aLegendFont[ 2 ], ::aTextColor, ::aBackColor, if( ::aLegendFont[ 3 ], BT_TEXT_BOLD, nil ), BT_TEXT_CENTER )
               nPrevRow := nRow - nBarHeight
               nPrevCol := nCol
            else // negative
               BT_DrawText ( ::hDC, nRow + nBarHeight + 3 + ::nPointSize / 2 , nCol, if( nData < 0, '-', '' ) + alltrim( transform( abs( nData ), ::cPicture ) ), ::aLegendFont[ 1 ], ::aLegendFont[ 2 ], ::aTextColor, ::aBackColor, if( ::aLegendFont[ 3 ], BT_TEXT_BOLD, nil ), BT_TEXT_CENTER )
               nPrevRow := nRow + nBarHeight
               nPrevCol := nCol
            endif
            nCol := nCol + nCategoryWidth
         next i
      next j
   endif
return nil


Method DrawScatterXYGraph Class GraphPlus
   local aScaleX := {}, aScaleY := {}
   local nMaxY, nMaxX, nMinY, nMinX, nUpperX, uUpperY, nLowerX, nLowerY, nRangeX, nRangeY, nLowerUpperRatioX, nLowerUpperRatioY, nLowerDivisionX, nLowerDivisionY, nDivisionX, nDivisionY, nDistanceX, nDistanceY, cDivisionX, cDivisionY, nDecimalLengthX, nDecimalLengthY, nZeroX, nZeroY, nZeroRowPos, nZeroColPos, nMaxWidthY, nScaleItems, nAvailableHeight, nTotalBarRatio, nScaleItemHeight, nHeight, nWidth, nDataX, nDataY, nMaxHeightX, nPrevRow, nPrevCol

   // draw title
   if hmg_len( ::cTitle ) > 0 
      aSize := BT_DrawTextSize (::hDC, ::cTitle, ::aTitleFont[ 1 ], ::aTitleFont[ 2 ], if( ::aTitleFont[ 3 ], BT_TEXT_BOLD, nil ))
      BT_DrawText ( ::hDC, ::nTop / 2, ::nWidth / 2, ::cTitle, ::aTitleFont[ 1 ], ::aTitleFont[ 2 ], ::aTitleColor, ::aBackColor, if( ::aTitleFont[ 3 ], BT_TEXT_BOLD, nil ), BT_TEXT_CENTER )
      ::nTop := ::nTop + aSize[ 2 ]
   endif
   ::DrawXYTitles()    // draw x/y titles
   if hmg_len( ::aData ) < 2
      msginfo( 'Not Enough data for XY!' )
      return nil
   endif
   if hmg_len( ::aData[ 1 ] ) <> hmg_len( ::aData[ 2 ] )
      msginfo( 'X and Y axis data should be in equal numbers!' )
      return nil
   endif
   // draw graph border
   && BT_DrawRectangle (::hDC, ::nTop, ::nLeft, ::nRight - ::nLeft , ::nBottom - ::nTop, ::aTextColor, 1 )
   // find max and min value for both x and y
   nMaxX := 0
   nMaxY := 0
   nMinX := 0
   nMinY := 0
   for i := 1 to hmg_len( ::aData[ 1 ] )
      nMaxX := Max( ::aData[ 1, i ], nMaxX )
      nMinX := Min( ::aData[ 1, i ], nMinX )
   next i   
      for i := 1 to hmg_len( ::aData[ 2 ] )
      nMaxY := Max( ::aData[ 2, i ], nMaxY )
      nMinY := Min( ::aData[ 2, i ], nMinY )
   next i   
   if nMaxX > 0
      nUpperX := nMaxX
   else 
      nUpperX := 0
   endif   
   if nMinX < 0
      nLowerX := nMinX
   else
      nLowerX := 0   
   endif
   if nMaxY > 0
      nUpperY := nMaxY
   else 
      nUpperY := 0
   endif   
   if nMinY < 0
      nLowerY := nMinY
   else
      nLowerY := 0   
   endif
   nRangeX := nUpperX - nLowerX
   nRangeY := nUpperY - nLowerY
   if nUpperX <> 0 .and. nLowerX <> 0
      if abs( nLowerX ) < nUpperX
         nLowerUpperRatioX := abs( nLowerX ) / nRangeX
         nLowerDivisionX := ::nVDivision * nLowerUpperRatioX
         if nLowerDivisionX - int( nLowerDivisionX ) > 0.0
            nLowerDivisionX := int( nLowerDivisionX ) + 1
         endif   
         nUpperDivisionX := ::nVDivision - nLowerDivisionX
         nDivisionX := nUpperX / nUpperDivisionX
      else
         nLowerUpperRatioX := abs( nUpperX ) / nRangeX
         nUpperDivisionX := ::nVDivision * nLowerUpperRatioX
         if nUpperDivisionX - int( nUpperDivisionX ) > 0.0
            nUpperDivisionX := int( nUpperDivisionX ) + 1
         endif   
         nLowerDivisionX := ::nVDivision - nUpperDivisionX
         nDivisionX := abs( nLowerX ) / nLowerDivisionX
      endif
   else
      nDivisionX := nRangeX / ::nVDivision
      if nUpperX > 0
         nUpperDivisionX := ::nVDivision
         nLowerDivisionX := 0
      else
         nLowerDivisionX := ::nVDivision
         nUpperDivisionX := 0
      endif         
   endif         
   if nUpperY <> 0 .and. nLowerY <> 0
      if abs( nLowerY ) < nUpperY
         nLowerUpperRatioY := abs( nLowerY ) / nRangeY
         nLowerDivisionY := ::nHDivision * nLowerUpperRatioY
         if nLowerDivisionY - int( nLowerDivisionY ) > 0.0
            nLowerDivisionY := int( nLowerDivisionY ) + 1
         endif   
         nUpperDivisionY := ::nHDivision - nLowerDivisionY
         nDivisionY := nUpperY / nUpperDivisionY
      else
         nLowerUpperRatioY := abs( nUpperY ) / nRangeY
         nUpperDivisionY := ::nHDivision * nLowerUpperRatioY
         if nUpperDivisionY - int( nUpperDivisionY ) > 0.0
            nUpperDivisionY := int( nUpperDivisionY ) + 1
         endif   
         nLowerDivisionY := ::nHDivision - nUpperDivisionY
         nDivisionY := abs( nLowerY ) / nLowerDivisionY
      endif
   else
      nDivisionY := nRangeY / ::nHDivision
      if nUpperY > 0
         nUpperDivisionY := ::nHDivision
         nLowerDivisionY := 0
      else
         nLowerDivisionX := ::nHDivision
         nUpperDivisionX := 0
      endif         
   endif        
   if int( nDivisionX ) > 0
      cDivisionX := alltrim( str( int( nDivisionX ), 14, 0 ) )
      if hmg_len( cDivisionX ) > 1 // if more than 1 digits round to power of 10s
         if substr( cDivisionX, 2, 1 ) <> '0'
            nDivisionX := ( val( left( cDivisionX, 2 ) ) + 1 ) * 10^ ( hmg_len( cDivisionX ) - 2 )
         endif   
      else
         if nDivisionX > int( nDivisionX ) 
            nDivisionX++
         endif
      endif
   else //
      cDivisionX := transform( nDivisionX, '9.9999999' )
      nDecimalLengthX := 8 - hmg_len( alltrim( str( val( right( cDivisionX, 7 ) ) ) ) )
      nDecimalX := val( right( cDivisionX, 7 ) )
      cDecimalX := alltrim( str( nDecimalX ) )
      nDivisionX := (val( alltrim( left( cDecimalX, 1 ) )) + 1 ) /(10 ^nDecimalLengthX)
   endif
   if int( nDivisionY ) > 0
      cDivisionY := alltrim( str( int( nDivisionY ), 14, 0 ) )
      if hmg_len( cDivisionY ) > 1 // if more than 1 digits round to power of 10s
         if substr( cDivisionY, 2, 1 ) <> '0'
            nDivisionY := ( val( left( cDivisionY, 2 ) ) + 1 ) * 10^ ( hmg_len( cDivisionY ) - 2 )
         endif   
      else
         if nDivisionY > int( nDivisionY ) 
            nDivisionY++
         endif
      endif
   else //
      cDivisionY := transform( nDivisionY, '9.9999999' )
      nDecimalLengthY := 8 - hmg_len( alltrim( str( val( right( cDivisionY, 7 ) ) ) ) )
      nDecimalY := val( right( cDivisionY, 7 ) )
      cDecimalY := alltrim( str( nDecimalY ) )
      nDivisionY := (val( alltrim( left( cDecimalY, 1 ) )) + 1 ) /(10 ^nDecimalLengthY)
   endif
   nRangeX := ::nVDivision * nDivisionX
   nRangeY := ::nHDivision * nDivisionY
   if hmg_len( ::cPictureX ) == 0 .and. hmg_len( ::cPictureY ) == 0 .and. hmg_len( ::cPicture ) > 0  // common picture defined by the user. 
      ::cPictureX := ::cPicture
      ::cPictureY := ::cPicture
   endif
   if hmg_len( ::cPictureX ) == 0
      ::cPictureX := replicate( '9', max( hmg_len( alltrim( str( nUpperX ) ) ), hmg_len( alltrim( str( nLowerX ) ) ) ) )
   endif   
   if hmg_len( ::cPictureY ) == 0
      ::cPictureY := replicate( '9', max( hmg_len( alltrim( str( nUpperX ) ) ), hmg_len( alltrim( str( nLowerX ) ) ) ) )
   endif   
   // find scales
   asize( aScaleX, 0 )
   asize( aScaleY, 0 )
   nDistanceX := nRangeX / ::nVDivision
   nDistanceY := nRangeY / ::nHDivision
   for i := nUpperDivisionX to 1 step -1
      aadd( aScaleX, alltrim( transform( nDistanceX * i, ::cPictureX ) ) )
   next i
   aadd( aScaleX, alltrim( transform( 0, ::cPictureX ) ) )
   nZeroX := hmg_len( aScaleX )
   for i := 1 to nLowerDivisionX
      aadd( aScaleX, '-' + alltrim( transform( nDistanceX * i, ::cPictureX ) ) )
   next i
   for i := nUpperDivisionY to 1 step -1
      aadd( aScaleY, alltrim( transform( nDistanceY * i, ::cPictureY ) ) )
   next i
   aadd( aScaleY, alltrim( transform( 0, ::cPictureY ) ) )
   nZeroY := hmg_len( aScaleY )
   for i := 1 to nLowerDivisionY
      aadd( aScaleY, '-' + alltrim( transform( nDistanceY * i, ::cPictureY ) ) )
   next i
   // find maxscaleitemwidthY
   nMaxWidthY := 0
   for i := 1 to hmg_len( aScaleY )
      aSize := BT_DrawTextSize (::hDC, aScaleY[ i ], ::aLegendFont[ 1 ], ::aLegendFont[ 2 ], if( ::aLegendFont[ 3 ], BT_TEXT_BOLD, nil ))
      nMaxWidthY := max( nMaxWidthY, aSize[ 1 ] )
   next i
   nMaxHeightX := 0
   for i := 1 to hmg_len( aScaleX )
      aSize := BT_DrawTextSize (::hDC, aScaleX[ i ], ::aLegendFont[ 1 ], ::aLegendFont[ 2 ], if( ::aLegendFont[ 3 ], BT_TEXT_BOLD, nil ))
      nMaxHeightX := max( nMaxHeightX, aSize[ 1 ] )
   next i

   // draw all items other than graph
   ::nTop := ::nTop + 20
   ::nBottom := ::nBottom - 20
   if ::lShowValues
      if nZeroX == hmg_len( aScaleX ) // X Zero at left
         ::nLeft := ::nLeft + nMaxWidthY
      endif
      if nZeroY == 1 // Y Zero at Bottom
         ::nBottom := ::nBottom - nMaxHeightX
      endif
   endif      
   
   ::nRight := ::nRight - 5
   ::nLeft := ::nLeft + 5
   
   // find zeroRow position
   do case
   case nZeroY == 1 // all are negative!
      nZeroRowPos := ::nTop
   case nZeroY == hmg_len( aScaleY ) // all are positive!
      nZeroRowPos := ::nBottom
   otherwise  // zero position in the middles
      nZeroRowPos := ::nTop + ( ( ::nBottom - ::nTop ) / ::nHDivision * ( nZeroY - 1 ) )
   endcase
   nZeroRowPos := int( nZeroRowPos )
   
   // find zeroCol position
   do case
   case nZeroX == 1 // all are negative!
      nZeroColPos := ::nRight
   case nZeroX == hmg_len( aScaleX ) // all are positive!
      nZeroColPos := ::nLeft
   otherwise  // zero position in the middles
      nZeroColPos := ::nRight - ( ( ::nRight - ::nLeft ) / ::nVDivision * ( nZeroX - 1 ) )
   endcase
   // Draw Grids
   if ::lHGrid 
      nHeight := ( ::nBottom - ::nTop ) / ::nHDivision
      nRow := ::nTop
      for i := 1 to hmg_len( aScaleY )
         BT_DrawLine (::hDC, nRow, ::nLeft, nRow, ::nRight, ::aGridLineColor, ::nLineWidth)
         nRow := nRow + nHeight
      next i
   endif
   
   if ::lVGrid 
      nWidth := ( ::nRight - ::nLeft ) / ::nVDivision
      BT_DrawLine (::hDC, ::nTop, ::nLeft, ::nBottom, ::nLeft, ::aGridLineColor, ::nLineWidth)
      nCol := ::nLeft
      for i := 1 to hmg_len( aScaleX )
         nCol := nCol + nWidth
         BT_DrawLine (::hDC, ::nTop, nCol, ::nBottom, nCol, ::aGridLineColor, ::nLineWidth)
      next i
   endif
   if ::lShowValues
      // draw X Axis Values
      nCol := ::nLeft
      nScaleItems := hmg_len( aScaleX )
      nAvailableWidth := ::nRight - ::nLeft 
      nTotalBarRatio := nScaleItems - 1 // equal space between categories
      nScaleItemWidth := nAvailableWidth / nTotalBarRatio // 100% width
      for i := hmg_len( aScaleX ) to 1 step -1
         BT_DrawText ( ::hDC, nZeroRowPos, nCol, aScaleX[ i ], ::aLegendFont[ 1 ], ::aLegendFont[ 2 ], ::aTextColor, ::aBackColor, if( ::aLegendFont[ 3 ], BT_TEXT_BOLD, nil ),  BT_TEXT_CENTER)
         nCol := nCol + nScaleItemWidth
      next i 
      // Draw Y Axis Values
      nScaleItems := hmg_len( aScaleY )
      nAvailableHeight := ::nBottom - ::nTop
      nTotalBarRatio := nScaleItems - 1 // equal space between scales
      nScaleItemHeight := nAvailableHeight / nTotalBarRatio 
      nRow := ::nTop
      for i := 1 to hmg_len( aScaleY )
         if i <> nZeroY
            BT_DrawText ( ::hDC, nRow + 2, nZeroColPos - 5, aScaleY[ i ], ::aLegendFont[ 1 ], ::aLegendFont[ 2 ], ::aTextColor, ::aBackColor, if( ::aLegendFont[ 3 ], BT_TEXT_BOLD, nil ),  BT_TEXT_BASELINE + BT_TEXT_RIGHT)
         endif   
         nRow := nRow + nScaleItemHeight
      next i
   endif
   for i := 1 to hmg_len( ::aData[ 1 ] )
      nDataX := ::aData[ 1, i ]
      nDataY := ::aData[ 2, i ]
      nWidth := int( ( ::nRight - ::nLeft ) * ( abs( nDataX ) / nRangeX ) )
      nHeight := int( ( ::nBottom - ::nTop ) * ( abs( nDataY ) / nRangeY ) )
      if nDataX > 0
         nCol := nZeroColPos + nWidth - ::nPointSize / 2
      else
         nCol := nZeroColPos - nWidth - ::nPointSize / 2
      endif
      if nDataY > 0
         nRow := nZeroRowPos - nHeight - ::nPointSize / 2
      else   
         nRow := nZeroRowPos + nHeight - ::nPointSize / 2
      endif   
      BT_DrawFillEllipse ( ::hDC, nRow, nCol, ::nPointSize, ::nPointSize, ::aColors[ 1 ], ::aColors[ 1 ], 1)
      if ::lDataLabels
         BT_DrawText ( ::hDC, nRow - 15 - ::nPointSize / 2 , nCol, '(' + alltrim( transform( nDataX, ::cPictureX ) ) + ', ' + alltrim( transform( nDataY, ::cPictureY ) ) + ')' , ::aLegendFont[ 1 ], ::aLegendFont[ 2 ], ::aTextColor, ::aBackColor, if( ::aLegendFont[ 3 ], BT_TEXT_BOLD, nil ), BT_TEXT_CENTER )
      endif
      if ::lScatterLine
         if i <> 1
            BT_DrawLine (::hDC, nPrevRow, nPrevCol, nRow + ::nPointSize / 2, nCol + ::nPointSize / 2, ::aColors[ 1 ], ::nLineWidth)              
         endif
      endif
      nPrevRow := nRow + ::nPointSize / 2
      nPrevCol := nCol + ::nPointSize / 2     
   next i
   
return nil


Method Save( cFileName ) Class GraphPlus
   if pcount() == 0
      cFileName := GetFile ()
   endif
   cFileName := alltrim( cFileName )
   if hmg_len( cFileName ) > 0   
      if file( cFileName ) 
         if .not. msgyesno( 'File already Exists! Do you want to overwrite?' )
            return nil
         endif
      endif
      BT_BitmapSaveFile( ::hBitMap, cFileName, BT_FILEFORMAT_PNG )
   endif      
return nil


Method DrawLegends Class GraphPlus
   local nMaxWidth := 0, i := 0, nMaxHeight := 0
   local nLegends := hmg_len( ::aLegends )
   local aSize := {}
   local nColumns := 0, nRows := 0
   local nRow := 0
   local nCol := 0
   for i := 1 to nLegends
      aSize := BT_DrawTextSize (::hDC, ::aLegends[ i ], ::aLegendFont[ 1 ], ::aLegendFont[ 2 ], if( ::aLegendFont[ 3 ], BT_TEXT_BOLD, nil ))
      nMaxWidth := max( nMaxWidth, aSize[ 1 ] )
      nMaxHeight := max( nMaxHeight, aSize[ 2 ] )
   next i   
   nMaxWidth := ( nMaxWidth + ::nLegendBoxSize + 10 ) // per legend
   nMaxHeight := nMaxHeight + 10                      // per legend
   ::nLegendWidth := nMaxWidth
   
   // find columns/rows
   if ::nLegendPos == 1 // bottom
      nColumns := int( ( ::nWidth -  20 ) / nMaxWidth )
      if nColumns == 0 // legend length is not fitting!
         nColumns := 1
      endif
      nColumns := min( nColumns, nLegends )
      nRows := int( nLegends / nColumns ) + ( if( mod( nLegends, nColumns ) > 0, 1, 0 ) )
   else // right
      nColumns := 1
      nRows := nLegends
      if nLegends * nMaxHeight > ( ::nBottom - ::nTop )
         // all the legends can not fit!
         nRows := int( ( ::nBottom - ::nTop ) / nMaxHeight )
      endif
   endif   
   // draw legends
   if ::nLegendPos == 1 // bottom
      ::nBottom := ::nBottom - ( nRows * nMaxHeight )
      nRow := ::nBottom + 5
      i := 1
      j := 1
      nCol := ( ::nWidth - nColumns * ::nLegendWidth ) / 2
      do while i <= nLegends
         if j > nColumns
            j := 1
            nRow := nRow + nMaxHeight
            nCol := ( ::nWidth - nColumns * ::nLegendWidth ) / 2
         endif
         BT_DrawFillRectangle ( ::hDC, nRow, nCol, ::nLegendBoxSize, max( nMaxHeight - 10, ::nLegendBoxSize ), ::aColors[ i ], ::aColors[ i ], 1 )
         BT_DrawText ( ::hDC, nRow, nCol + ::nLegendBoxSize + 5, ::aLegends[ i ], ::aLegendFont[ 1 ], ::aLegendFont[ 2 ], ::aTextColor, ::aBackColor, if( ::aLegendFont[ 3 ], BT_TEXT_BOLD, nil ))
         nCol := nCol + nMaxWidth
         j++
         i++
      enddo
   else  // right
      ::nRight := ::nRight - nMaxWidth
      nRow := ::nTop + 5
      nCol := ::nRight + 5
      for i := 1 to nRows
         BT_DrawFillRectangle ( ::hDC, nRow, nCol, ::nLegendBoxSize, max( nMaxHeight - 10, ::nLegendBoxSize ), ::aColors[ i ], ::aColors[ i ], 1 )
         BT_DrawText ( ::hDC, nRow, nCol + ::nLegendBoxSize + 5, ::aLegends[ i ], ::aLegendFont[ 1 ], ::aLegendFont[ 2 ], ::aTextColor, ::aBackColor, if( ::aLegendFont[ 3 ], BT_TEXT_BOLD, nil ))
         nRow := nRow + nMaxHeight         
      next i   
   endif   
return nil

Method DrawSunBurstLegends( aLegends, aColors ) Class GraphPlus
   local nMaxWidth := 0, i := 0, nMaxHeight := 0
   local nLegends := hmg_len( aLegends )
   local aSize := {}
   local nColumns := 0, nRows := 0
   local nRow := 0
   local nCol := 0
   for i := 1 to nLegends
      aSize := BT_DrawTextSize (::hDC, aLegends[ i ], ::aLegendFont[ 1 ], ::aLegendFont[ 2 ], if( ::aLegendFont[ 3 ], BT_TEXT_BOLD, nil ))
      nMaxWidth := max( nMaxWidth, aSize[ 1 ] )
      nMaxHeight := max( nMaxHeight, aSize[ 2 ] )
   next i   
   nMaxWidth := ( nMaxWidth + ::nLegendBoxSize + 10 ) // per legend
   nMaxHeight := nMaxHeight + 10                      // per legend
   ::nLegendWidth := nMaxWidth
   
   // find columns/rows
   if ::nLegendPos == 1 // bottom
      nColumns := int( ( ::nWidth -  20 ) / nMaxWidth )
      if nColumns == 0 // legend length is not fitting!
         nColumns := 1
      endif
      nColumns := min( nColumns, nLegends )
      nRows := int( nLegends / nColumns ) + ( if( mod( nLegends, nColumns ) > 0, 1, 0 ) )
   else // right
      nColumns := 1
      nRows := nLegends
      if nLegends * nMaxHeight > ( ::nBottom - ::nTop )
         // all the legends can not fit!
         nRows := int( ( ::nBottom - ::nTop ) / nMaxHeight )
      endif
   endif   
   // draw legends
   if ::nLegendPos == 1 // bottom
      ::nBottom := ::nBottom - ( nRows * nMaxHeight )
      nRow := ::nBottom + 5
      i := 1
      j := 1
      nCol := ( ::nWidth - nColumns * ::nLegendWidth ) / 2
      do while i <= nLegends
         if j > nColumns
            j := 1
            nRow := nRow + nMaxHeight
            nCol := ( ::nWidth - nColumns * ::nLegendWidth ) / 2
         endif
         BT_DrawFillRectangle ( ::hDC, nRow, nCol, ::nLegendBoxSize, max( nMaxHeight - 10, ::nLegendBoxSize ), aColors[ i ], aColors[ i ], 1 )
         BT_DrawText ( ::hDC, nRow, nCol + ::nLegendBoxSize + 5, aLegends[ i ], ::aLegendFont[ 1 ], ::aLegendFont[ 2 ], ::aTextColor, ::aBackColor, if( ::aLegendFont[ 3 ], BT_TEXT_BOLD, nil ))
         nCol := nCol + nMaxWidth
         j++
         i++
      enddo
   else  // right
      ::nRight := ::nRight - nMaxWidth
      nRow := ::nTop + 5
      nCol := ::nRight + 5
      for i := 1 to nRows
         BT_DrawFillRectangle ( ::hDC, nRow, nCol, ::nLegendBoxSize, max( nMaxHeight - 10, ::nLegendBoxSize ), aColors[ i ], aColors[ i ], 1 )
         BT_DrawText ( ::hDC, nRow, nCol + ::nLegendBoxSize + 5, aLegends[ i ], ::aLegendFont[ 1 ], ::aLegendFont[ 2 ], ::aTextColor, ::aBackColor, if( ::aLegendFont[ 3 ], BT_TEXT_BOLD, nil ))
         nRow := nRow + nMaxHeight         
      next i   
   endif   
return nil


Method DrawCategories Class GraphPlus
   local nCategories := 0
   local nRow := 0, nCol := 0, nData := 0
   local nTotalBars := 0
   local nAvailableWidth := 0, nAvailableHeight := 0
   local nTotalBarRatio := 0
   local nCategoryRatio := 0
   local nTotalBarSize := 0
   local aSize := {}, nMaxWidth := 0

   local nBarWidth := 0, nBarHeight := 0
   local nCategoryWidth := 0
   local nFactor := 0
   local nBarGap := 0
   nCategories := hmg_len( ::aCategories )
   if nCategories == 0
      return nil
   endif
   if .not. ::lGraphRotated // columns and like
      nAvailableWidth := ::nRight - ( ::nLeft + ::nMaxScaleWidth + 5 )
      if ::nGraphType <> GT_AREA  // area charts start from left and ends at right. 
         nTotalBarRatio := nCategories // equal space between categories
         nCategoryWidth := nAvailableWidth / nTotalBarRatio // 100% width
         nCol := ::nLeft + ::nMaxScaleWidth + 5 + ( nCategoryWidth / 2 )
      else   
         nTotalBarRatio := nCategories - 1// equal space between categories
         nCategoryWidth := nAvailableWidth / nTotalBarRatio // 100% width
         nCol := ::nLeft + ::nMaxScaleWidth + 5
      endif   
      nRow := ::nZeroPos
      nMaxHeight := 0
      asize := {}
      for i := 1 to hmg_len( ::aCategories )
         aSize := BT_DrawTextSize (::hDC, ::aCategories[ i ], ::aLegendFont[ 1 ], ::aLegendFont[ 2 ], if( ::aLegendFont[ 3 ], BT_TEXT_BOLD, nil ))
         nMaxHeight := max( nMaxHeight, aSize[ 2 ] )
      next i
      for i := 1 to hmg_len( ::aCategories )
         BT_DrawText ( ::hDC, ::nBottom, nCol, ::aCategories[ i ], ::aLegendFont[ 1 ], ::aLegendFont[ 2 ], ::aTextColor, ::aBackColor, if( ::aLegendFont[ 3 ], BT_TEXT_BOLD, nil ),  BT_TEXT_CENTER)
         nCol := nCol + nCategoryWidth
      next i
      ::nBottom := ::nBottom - nMaxHeight
   else
      nAvailableHeight := ::nBottom - ::nTop
      nTotalBarRatio := nCategories // equal space between categories

      nBarHeight := nAvailableHeight / nTotalBarRatio // 100% width
      nCategoryHeight := nBarHeight
      
      nRow := ::nTop + ( nCategoryHeight / 2 )
      nMaxWidth := 0
      asize := {}
      for i := 1 to hmg_len( ::aCategories )
         aSize := BT_DrawTextSize (::hDC, ::aCategories[ i ], ::aLegendFont[ 1 ], ::aLegendFont[ 2 ], if( ::aLegendFont[ 3 ], BT_TEXT_BOLD, nil ))
         nMaxWidth := max( nMaxWidth, aSize[ 1 ] )
      next i
      nCol := ::nLeft + nMaxWidth
      for i := 1 to hmg_len( ::aCategories )
         BT_DrawText ( ::hDC, nRow + 2, nCol, ::aCategories[ i ], ::aLegendFont[ 1 ], ::aLegendFont[ 2 ], ::aTextColor, ::aBackColor, if( ::aLegendFont[ 3 ], BT_TEXT_BOLD, nil ), BT_TEXT_BASELINE + BT_TEXT_RIGHT)
         nRow := nRow + nCategoryHeight
      next i
      ::nLeft := ::nLeft + nMaxWidth
      ::nLeft := ::nLeft + 5
   endif
return nil

Method DrawXYTitles Class GraphPlus
   local aSize := {}
   local cXTitle := ::cXTitle
   local cYTitle := ::cYTitle
   if ::lGraphRotated // bar graph -> Swap x and y
      cYTitle := ::cXTitle
      cXTitle := ::cYTitle
   endif      
   if hmg_len( cXTitle ) > 0
      aSize := BT_DrawTextSize (::hDC, cXTitle, ::aTitleFont[ 1 ], ::aTitleFont[ 2 ] - 2 , if( ::aTitleFont[ 3 ], BT_TEXT_BOLD, nil ))
      BT_DrawText (::hDC, ::nBottom - aSize[ 2 ], (::nRight - ::nLeft) / 2, cXTitle, ::aTitleFont[ 1 ], ::aTitleFont[ 2 ] - 2, ::aTitleColor, ::aBackColor, if( ::aTitleFont[ 3 ], BT_TEXT_BOLD, nil ), BT_TEXT_CENTER)
      ::nBottom := ::nBottom - aSize[ 2 ]
   endif
   if hmg_len( cYTitle ) > 0
      aSize := BT_DrawTextSize (::hDC, cYTitle, ::aTitleFont[ 1 ], ::aTitleFont[ 2 ] - 2 , if( ::aTitleFont[ 3 ], BT_TEXT_BOLD, nil ))
      BT_DrawText (::hDC, (::nBottom - ::nTop) / 2, ::nLeft, cYTitle, ::aTitleFont[ 1 ], ::aTitleFont[ 2 ] - 2, ::aTitleColor, ::aBackColor, if( ::aTitleFont[ 3 ], BT_TEXT_BOLD, nil ), BT_TEXT_CENTER, BT_TEXT_VERTICAL_ASCENDANT)
      ::nLeft := ::nLeft + aSize[ 2 ]
   endif
return nil

Method InitColors Class GraphPlus
   local i, j, nLegends, nFactor, nCurFactor := 5
   local aHSL
   local aColorTheme
   nLegends := hmg_len( ::aLegends )
   if hmg_len( ::aColors ) = 0 .or. hmg_len( ::aColors ) < nLegends
      asize( ::aColors, 0 )
      aColorTheme := ::aColorThemes[ ::nColorTheme ]
      if nLegends <= hmg_len( aColorTheme )
         for i := 1 to nLegends
            aadd( ::aColors, aColorTheme[ i ] )
         next i
      else
         for i := 1 to hmg_len( aColorTheme )
            aadd( ::aColors, aColorTheme[ i ] )         
         next i
         i := hmg_len( aColorTheme ) + 1
         j := 1
         nFactor := 5
         do while i <= hmg_len( ::aLegends )
            if j > hmg_len( aColorTheme )
               j := 1
               nCurFactor *= nFactor
            endif   
            aHSL := GP_RGB2HSL( aColorTheme[ j ] )
            aHSL[ 3 ] := aHSL[ 3 ] + iif( ::lLighter, nCurFactor, -nCurFactor )
            aadd( ::aColors, GP_HSL2RGB( aHSL ) )
            j++
            i++
         enddo
      endif
   endif
return nil

Method SwitchRowsToColumns
   local aTemp := {}, aLine := {}
   local cTemp := ''
   local i := 0
   local j := 0
   // transpose data
   for i := 1 to hmg_len( ::aData[ 1 ] )
      asize( aLine, 0 )
      for j := 1 to hmg_len( ::aData )
         aadd( aLine, ::aData[ j, i ] )
      next j
      aadd( aTemp, aclone( aLine ) )
   next i
   asize( ::aData, 0 )   
   ::aData := aclone( aTemp )
   // switch legends and categories
   aTemp := aclone( ::aCategories )
   ::aCategories := aclone( ::aLegends )
   ::aLegends := aclone( aTemp )
   // switch xy titles
   cTemp := ::cXTitle
   ::cXTitle := ::cYTitle
   ::cYTitle := cTemp
return nil   

Method FindPieCoordinates( nChartSize ) Class GraphPlus
   local aPoints := array( 14 )
   local fromcol, tocol, torow, fromrow
   afill( aPoints, 0 )
   fromcol := ::nLeft + ( ( ::nRight - ::nLeft ) - nChartSize ) / 2 
   tocol := ::nRight - ( ( ::nRight - ::nLeft ) - nChartSize ) / 2 
   torow := ::nBottom - ( ( ::nBottom - ::nTop ) - nChartSize ) / 2
   fromrow := ::nTop + ( ( ::nBottom - ::nTop ) - nChartSize ) / 2
   aPoints[ 1 ] := fromrow // topleftrow
   aPoints[ 2 ] := fromcol // topleftcol
   aPoints[ 3 ] := fromrow // toprightrow
   aPoints[ 4 ] := tocol // toprightcol
   aPoints[ 5 ] := torow // bottomrightrow
   aPoints[ 6 ] := tocol // bottomrightcol
   aPoints[ 7 ] := torow // bottomleftrow
   aPoints[ 8 ] := fromcol // bottomleftcol
   aPoints[ 9 ] := fromcol + int(tocol - fromcol) / 2 // middletopcol
   aPoints[ 10 ] := fromrow + int(torow - fromrow) / 2 // middleleftrow
   aPoints[ 11 ] := fromcol // middleleftcol
   aPoints[ 12 ] := fromcol + int(tocol - fromcol) / 2 // middlebottomcol
   aPoints[ 13 ] := fromrow + int(torow - fromrow) / 2 // middlerightrow
   aPoints[ 14 ] := tocol // middlerightcol
return aClone( aPoints )

Method FillSunBurstData() Class GraphPlus
   local i := 0
   local cCurrentHeader := ::aData[ 1, 1 ]
   local nColorID := 1
   local nHeadTotal := 0
   local aHeadData := {}
   local aCurTotal := array( ::nLevels )
   local aCurHeader := {}
   local aConcatData := aclone( ::aData )
   afill( aCurTotal, 0 )
   asize( ::aLegends, 0 )
   asize( ::aLevelHeaders, 0 )
   asize( ::aLevelValues, 0 )
   for i := 1 to hmg_len( aConcatData )
      for j := 1 to ::nLevels
         cHeader := ''
         for k := 1 to j
            cHeader := cHeader + if( hmg_len( alltrim( cHeader ) ) > 0, ': ', '' ) +  ::aData[ i, k ]
         next k
         aConcatData[ i, j ] := cHeader
      next j
   next i
   for i := 1 to ::nLevels
      aadd( aCurHeader, aConcatData[ 1, i ] )
      aadd( ::aLevelHeaders, {} )
      aadd( ::aLevelValues, {} )
   next i   
   for i := 1 to hmg_len( aConcatData )
      for j := ::nLevels to 1 step -1
         if aConcatData[ i, j ] <> aCurHeader[ j ]
            aadd( ::aLevelValues[ j ], { nColorID, aCurTotal[ j ] } )
            aadd( ::aLevelHeaders[ j ], ::aData[ i-1, j ] ) //aCurHeader[ j ] )
            if j == 1
               nColorID := nColorID + 1
               aadd( ::aLegends, aCurHeader[ j ] )
            endif            
            aCurHeader[ j ] := aConcatData[ i, j ]
            aCurTotal[ j ] := aConcatData[ i, ::nLevels + 1 ]
         else
            aCurTotal[ j ] := aCurTotal[ j ] + aConcatData[ i, ::nLevels + 1 ]         
         endif   
      next j
   next i
   // last row!
   for j := 1 to ::nLevels
      aadd( ::aLevelValues[ j ], { nColorID, aCurTotal[ j ] } )
      aadd( ::aLevelHeaders[ j ], ::aData[ i - 1, j ] )
   next j      
   aadd( ::aLegends, aCurHeader[ 1 ] )
return nil

Method FillTreeMapData() Class GraphPlus
   local i := 0
   local cCurrentHeader := ::aData[ 1, 1 ]
   local nColorID := 1
   local nHeadTotal := 0
   local aHeadData := {}
   local aCurTotal := array( ::nLevels )
   local aCurHeader := {}
   local aConcatData := aclone( ::aData )
   afill( aCurTotal, 0 )
   asize( ::aLegends, 0 )
   asize( ::aLevelHeaders, 0 )
   asize( ::aLevelValues, 0 )
   for i := 1 to hmg_len( aConcatData )
      for j := 1 to ::nLevels
         cHeader := ''
         for k := 1 to j
            cHeader := cHeader + if( hmg_len( alltrim( cHeader ) ) > 0, ': ', '' ) + ::aData[ i, k ]
         next k
         aConcatData[ i, j ] := cHeader
      next j
   next i
   for i := 1 to ::nLevels
      aadd( aCurHeader, aConcatData[ 1, i ] )
      aadd( ::aLevelHeaders, {} )
      aadd( ::aLevelValues, {} )
   next i   
   for i := 1 to hmg_len( aConcatData )
      for j := ::nLevels to 1 step -1
         if aConcatData[ i, j ] <> aCurHeader[ j ]
            aadd( ::aLevelValues[ j ], { nColorID, aCurTotal[ j ], aCurHeader[ j ], if( j > 1, aCurHeader[ j - 1 ], '' ) } )
            aadd( ::aLevelHeaders[ j ], ::aData[ i-1, j ] ) //aCurHeader[ j ] )
            if j == 1
               nColorID := nColorID + 1
               aadd( ::aLegends, aCurHeader[ j ] )
            endif            
            aCurHeader[ j ] := aConcatData[ i, j ]
            aCurTotal[ j ] := aConcatData[ i, ::nLevels + 1 ]
         else
            aCurTotal[ j ] := aCurTotal[ j ] + aConcatData[ i, ::nLevels + 1 ]         
         endif   
      next j
   next i
   // last row!
   for j := 1 to ::nLevels
      aadd( ::aLevelValues[ j ], { nColorID, aCurTotal[ j ], aCurHeader[ j ], if( j > 1, aCurHeader[ j - 1 ], '' ) } )
      aadd( ::aLevelHeaders[ j ], ::aData[ i - 1, j ] )
   next j      
   aadd( ::aLegends, aCurHeader[ 1 ] )
return nil


/**
 * Copyright (C) 2001 by University of Maryland, College Park, MD 20742, USA 
 * and Martin Wattenberg, w@bewitched.com
 * All rights reserved.
 * Authors: Benjamin B. Bederson and Martin Wattenberg
 * http://www.cs.umd.edu/hcil/treemaps
 */

Method StripTreemap( nRow1, nCol1, nWidth1, nHeight1 ) Class GraphPlus
   local lDebug := .f.
   local aLayoutBox := { nRow1, nCol1, nWidth1, nHeight1 } // rectangle with 4 items row, col, width and height
   local aBounds := aclone( aLayoutBox )
   
   local lLookAhead := .t.
	local i := 0
	local ntotalSize := 0
   local nArea := 0, nScaleFactor := 0
  	local nFinishedIndex := 1
	local nNumItems := 0
	local nPrevAR := 0
	local nAR := 0
	local nWidth := 0
	local nRowOffset := 0
   local aBox := {}

   for i := 1 to hmg_len( ::aTreeMapData )
      nTotalSize := nTotalSize + ::aTreeMapData[ i ]
   next i   
   nArea := aLayoutBox[ 3 ] * aLayoutBox[ 4 ]
   nScaleFactor := ( nArea/nTotalSize ) ^ 0.5
   aBox := aclone( aLayoutBox )
   aBox[ 1 ] := aBox[ 1 ] / nScaleFactor
   aBox[ 2 ] := aBox[ 2 ] / nScaleFactor
   aBox[ 3 ] := aBox[ 3 ] / nScaleFactor
   aBox[ 4 ] := aBox[ 4 ] / nScaleFactor
   do while nFinishedIndex <= hmg_len( ::aTreeMapData )
      // Layout strip
      nNumItems = ::LayoutStrip( aBox, nFinishedIndex )
    	// Lookahead to second strip
      if lLookAhead
         if ((nFinishedIndex + nNumItems) <= hmg_len( ::aTreeMapData ) )
    		   nNumItems2 := 0
		      nAR2a := 0
		      nAR2b := 0
				// Layout 2nd strip and compute AR of first strip plus 2nd strip
		      nNumItems2 = ::LayoutStrip(aBox, nFinishedIndex + nNumItems)
            nAR2a = ::ComputeAverageAspectRatio(nFinishedIndex, nNumItems + nNumItems2)
				// Layout 1st and 2nd strips together
		      ::ComputeHorizontalBoxLayout(aBox, nFinishedIndex, nNumItems + nNumItems2)
		      nAR2b = ::ComputeAverageAspectRatio(nFinishedIndex, nNumItems + nNumItems2)
//		      msginfo("F: numItems2 = " + str( nNumItems2 ) + ", ar2a="+ str( nAR2a ) +", ar2b="+str( nAR2b) )
            if nAR2b < nAR2a 
			      nNumItems := nNumItems + nNumItems2
//			      msginfo("G: numItems = " + str( nNumItems) )
		      else
			      ::computeHorizontalBoxLayout(aBox, nFinishedIndex, nNumItems)
//			      msginfo("H: backup numItems = " + str(nNumItems))
		      endif
         endif
      endif 
      for i := nFinishedIndex to nFinishedIndex + nNumItems - 1
         ::aTreeMapRect[ i, 1 ] := ::aTreeMapRect[ i, 1 ] + nRowOffset
      next i   
      nHeight := ::aTreeMapRect[ nFinishedIndex, 4 ]
	   nRowoffset := nRowOffset + nHeight
      aBox[ 1 ] := aBox[ 1 ] + nHeight
      aBox[ 4 ] := aBox[ 4 ] - nHeight
	   nFinishedIndex := nFinishedIndex + nNumItems
   enddo   
   aRect := { }
   for i := 1 to hmg_len( ::aTreeMapData )
      aRect := ::aTreeMapRect[ i ]
      aRect[ 1 ] := aRect[ 1 ] * nScaleFactor
      aRect[ 2 ] := aRect[ 2 ] * nScaleFactor
      aRect[ 3 ] := aRect[ 3 ] * nScaleFactor
      aRect[ 4 ] := aRect[ 4 ] * nScaleFactor
      aRect[ 1 ] := aRect[ 1 ] + aBounds[ 1 ]
      aRect[ 2 ] := aRect[ 2 ] + aBounds[ 2 ]
      ::aTreeMapRect[ i ] := aClone( aRect )
   next i   
return nil

Method LayoutStrip( aBox, nIndex ) Class GraphPlus
	local nNumItems := 0
   local nPrevAR := 0
   local nAR := 2^20
   local nWidth := 0
   nPrevAr := nAR
   nNumItems := nNumItems + 1
   nWidth := ::ComputeHorizontalBoxLayout(aBox, nIndex, nNumItems)
   nAR := ::ComputeAverageAspectRatio(nIndex, nNumItems)
   do while (nAR < nPrevAR) .and. ((nIndex + nNumItems) <= hmg_len( ::aTreeMapData ) )
      nPrevAr := nAR
      nNumItems := nNumItems + 1
      nWidth := ::ComputeHorizontalBoxLayout(aBox, nIndex, nNumItems)
      nAR := ::ComputeAverageAspectRatio(nIndex, nNumItems)
   enddo      
   if nAR >= nPrevAR
      nNumItems := nNumItems - 1 
	   nWidth = ::ComputeHorizontalBoxLayout(aBox, nIndex, nNumItems)
	   nAR = ::ComputeAverageAspectRatio(nIndex, nNumItems)
//	    debug("L.2: backup: numItems="+numItems);
	endif
return nNumItems

Method ComputeHorizontalBoxLayout(aBox, nIndex, nNumItems) Class GraphPlus
   local i := 0
	local nTotalSize := ::ComputeSize(nIndex, nNumItems)
	local nHeight := nTotalSize / aBox[ 3 ]
	local nWidth := 0
	local nCol := 0
   for i := nIndex to nIndex + nNumItems - 1
      nWidth := ::aTreeMapData[ i ] / nHeight
      ::aTreeMapRect[ i ] := { 0, nCol, nWidth, nHeight }
	   nCol := nCol + nWidth
   next i
return nWidth

Method ComputeSize( nIndex, nNum ) Class GraphPlus
   local nSize := 0
   local i := 0
   for i := nIndex to nIndex + nNum - 1
//      if i + nIndex <= hmg_len( ::aTreeMapData )
         nSize := nSize + ::aTreeMapData[ i ] 
//      endif   
   next i
return nSize

Method ComputeAverageAspectRatio(nIndex, nNumItems) Class GraphPlus
	local nAR := 0
	local nTAR := 0
   local nW := 0
   local nH := 0
   local i := 0
   for i := nIndex to nIndex + nNumItems - 1
      nW := ::aTreeMapRect[ i, 3 ]
      nH := ::aTreeMapRect[ i, 4 ]
      nAR := max( nW/ nH, nH / nW )
      nTAR := nTAR + nAR
   next i   
	nTAR := nTAR / nNumItems
return nTAR


Method ComputeAspectRatio( nIndex) Class GraphPlus
   local nW := ::aTreeMapRect[ 1, nIndex ][ 3 ] 
   local nH := ::aTreeMapRect[ 1, nIndex ][ 4 ] 
   local nAR := max( nW / nH, nH / nW )
return nAR


function findboundary( nNum )
   LOCAL nE, nMax, nMan, nVal, nOffset

   nE:=10
   nVal:=0
   nNum:=Abs(nNum)

   DO WHILE .T.

      nMax := 10**nE

      IF Int(nNum/nMax)>0

         nMan:=(nNum/nMax)-Int(nNum/nMax)
         nOffset:=1
         nOffset:=IF(nMan<=.75,.75,nOffset)
         nOffset:=IF(nMan<=.50,.50,nOffset)
         nOffset:=IF(nMan<=.25,.25,nOffset)
         nOffset:=IF(nMan<=.00,.00,nOffset)
         nVal := (Int(nNum/nMax)+nOffset)*nMax
         EXIT

      ENDIF

      nE--

   ENDDO

RETURN (nVal)
   

function GetNegativeColor( aRGB )
   local aNewRGB := {}
   aNewRGB := { 255 - aRGB[ 1 ], 255 - aRGB[ 2 ], 255 - aRGB[ 3 ] }
return aNewRGB
   
FUNCTION GP_RGB2HSL( aRGB )

   LOCAL nMax, nMin, nDelta
   LOCAL nH, nS, nL, nR, nG, nB

   nR := aRGB[1] / 255
   nG := aRGB[2] / 255
   nB := aRGB[3] / 255
   nMax := Max( nR, Max( nG, nB ) )
   nMin := Min( nR, Min( nG, nB ) )
   nDelta := nMax - nMin

   IF nDelta == 0
      nH := 0
   ELSEIF ( nMax == nR )
      nH := (( nG - nB) / nDelta) % 6
   ELSEIF ( nMax == nG )
      nH := (nB - nR) / nDelta + 2
   ELSE
      nH := (nR - nG) / nDelta + 4
   ENDIF
   nH := Int( nH * 60 )
   IF ( nH < 0 )
      nH += 360
   ENDIF
   nL := ( nMax + nMin ) / 2
   nS := iif( nDelta == 0, 0, nDelta / ( 1 - Abs( 2 * nL - 1 )) )
   nS := nS * 100
   nL := nL * 100
   RETURN { nH, nS, nL }

FUNCTION GP_HSL2RGB( aHSL )

   LOCAL nR := 0, nG := 0, nB := 0, nH, nS, nL, c, x, m
   nH := aHSL[1]
   nS := aHSL[2] / 100
   nL := aHSL[3] / 100
   c := ( 1 - Abs( 2 * nL - 1 ) ) * nS
   x := c * ( 1 - Abs( ( nH / 60 ) % 2 - 1 ) )
   m := nL - c / 2
   IF (0 <= nH .And. nH < 60)
      nR := c
      nG := x
      nB := 0
   ELSEIF (60 <= nH .And. nH < 120)
      nR := x
      nG := c
      nB := 0
   ELSEIF (120 <= nH .And. nH < 180)
      nR := 0
      nG := c
      nB := x
   ELSEIF (180 <= nH .And. nH < 240)
      nR := 0
      nG := x
      nB := c
   ELSEIF (240 <= nH .And. nH < 300)
      nR := x
      nG := 0
      nB := c
   ELSEIF (300 <= nH .And. nH < 360)
      nR := c
      nG := 0
      nB := x
   ENDIF
   nR := Int( ( nR + m ) * 255 )
   nG := Int( ( nG + m ) * 255 )
   nB := Int( ( nB + m ) * 255 )
   RETURN { nR, nG, nB }   
   
function degreetoradian( nDegree )
   return ( nDegree * 0.0174532925 )

function radiantodegree( nRadian )
   return ( nRadian / 0.0174532925 )  

